/********************************************************************
 * @file    cmand.c
 * @brief   Data reporting and downloading functions
 * @author  $LastChangedBy: rcraven $ (hiccups@cmand.org)
 * @version $LastChangedRevision: 546 $
 * @date    $Date: 2014-10-07 08:10:28 -0700 (Tue, 07 Oct 2014) $
 *
 * $Id: cmand.c 546 2014-10-07 15:10:28Z rcraven $
 ********************************************************************/
#include "hiccups.h"
#include "settings.h"

/**
 * Converts a time string from the CMAND webserver to a time struct.
 * The format will be YYYY-MM-DDTHH:MM:SS.
 *
 * @param tstr The time string to convert
 * @param t The time struct to populate
 */
static void strtotime(char *tstr, struct tm *t)
{
    /* break up tstr by changing separators to null chars */
    tstr[4] = '\0';
    tstr[7] = '\0';
    tstr[10] = '\0';
    tstr[13] = '\0';
    tstr[16] = '\0';

    /* each broken up piece of tstr is a time field */
    t->tm_sec = atoi(tstr + 17);
    t->tm_min = atoi(tstr + 14);
    t->tm_hour = atoi(tstr + 11);
    t->tm_mday = atoi(tstr + 8);
    t->tm_mon = atoi(tstr + 5) - 1; /* num months since Jan */
    t->tm_year = atoi(tstr) - 1900; /* num years since 1900 */
}

/**
 * Opens an HTTP connection to the CMAND webserver
 *
 * @return The opened socket
 */
static int connect_to_cmand(void)
{
    struct sockaddr_in dst;
    int sock, err;

    memset(&dst, 0, sizeof dst);
    dst.sin_family = AF_INET;
    dst.sin_port = htons(80);
    if (batch_mode)
        inet_aton(CMAND_IP, &dst.sin_addr);
    else
        resolve_to_ip(CMAND_URL, &dst);

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) return -1;

    err = connect(sock, (const struct sockaddr *) &dst, sizeof (dst));
    if (err < 0) return -1;

    return sock;
}

/**
 * Execute an HTTP GET of the given file and return the response
 *
 * @param file The name of the file to download (leave off /hiccups)
 * @param msg The response from the webserver with HTTP headers stripped
 * @return the number of bytes in the response, or -1 on error
 */
static int get_from_cmand(char *file, char **msg)
{
    char *req, *resp;
    char *req_skel = "GET /%s%s HTTP/1.1\r\n"
                     "Host: %s\r\n"
                     "User-Agent: %s\r\n"
                     "Pragma: no-cache\r\n"
                     "Cache-Control: no-cache\r\n\r\n";
    int sock, err, bytes_read = -1;

    req = malloc(strlen(req_skel) + strlen(CMAND_PREFIX) + strlen(file) + 
                 strlen(CMAND_URL) + strlen(PACKAGE_STRING));
    sprintf(req, req_skel, CMAND_PREFIX, file, CMAND_URL, PACKAGE_STRING);

    sock = connect_to_cmand();
    if (sock < 0) goto cleanbuf;

    /* send the GET request */
    err = send(sock, req, strlen(req), 0);
    if (err < 0) goto cleanall;

    /* read the response */
    resp = calloc(BUFSIZE*4, sizeof (char));
    bytes_read = recv_tmo(sock, resp, BUFSIZE*4, CMAND_TMO);

    if (bytes_read > 0) {
        /* extract the actual message from the HTTP headers */
        resp[bytes_read] = '\0';
        *msg = strdup(trim(strstr(trim(resp), "\r\n\r\n")));
    }
    free(resp);

cleanall:
    net_close(sock);
cleanbuf:
    free(req);
    return bytes_read;
}

/**
 * Queries the CMAND webserver for the current UTC time
 *
 * The goal is to make a best effort attempt to read the time from
 * the server.  If we cannot, just bail and the system time will
 * be used without an offset.
 *
 * @return the time in seconds since the Unix epoch
 */
static int get_network_time(void)
{
    char *tstr;
    struct tm t;
    int len;

    /* download the server time */
    len = get_from_cmand(CMAND_PATH_TIME, &tstr);
    if (len <= 0) return 0;

    /* parse it */
    memset(&t, 0, sizeof (t));
    strtotime(tstr, &t);
    free(tstr);
    
    /* convert to seconds since epoch */
    return mktime(&t);
}

/**
 * Compares our local version with the version posted on the web
 *
 * Queries the CMAND webserver for the latest version and prints a 
 * message to the user if they're out of date.
 */
void check_for_new_version(void)
{
    char *vstr;
    int len;

    /* download the latest version */
    len = get_from_cmand(CMAND_PATH_VERSION, &vstr);
    if (len <= 0) { warn("Version check failed"); return; }

    if (strcmp(VERSION, vstr) != 0) {
        color_printf(TERM_WHITE, "WARNING: ");
        printf("You are not using the latest version of this software.  You\n");
        printf("         should upgrade to ensure you are getting accurate results.\n");
        printf("         Your version:   %s\n", VERSION);
        printf("         Latest version: ");
        color_printf(TERM_WHITE, "%s", vstr);
        printf("\n\n");
    }
    free(vstr);
}

/**
 * Downloads the server list posted on the web.
 *
 * @param The buffer to store the list
 * @return The number of targets downloaded
 */
int download_servlist(char **targlist)
{
    int len, i, num_targs = 0;

    len = get_from_cmand(CMAND_PATH_SERVLIST, targlist);
    if (len <= 0) return 0;

    /* count up the number of targets in the list */
    for (i = 0; (*targlist)[i]; i++) {
        if ((*targlist)[i] == '\n') num_targs++;
    }
    num_targs++; /* str is always trimmed, so add one for last line */
    return num_targs;
}

/**
 * Initialize the CMAND upload pack.
 *
 * @param pk The upload pack to initialize
 */
void pk_init(struct upload_pack *pk)
{
    time_t tm, system_utc, network_utc;

    pk->str = malloc(BUFSIZE*2);
    pk->ready = BUFSIZE*2;
    pk->used = 0;

    tm = time(NULL);
    system_utc = mktime(gmtime(&tm));
    network_utc = get_network_time();
    if (network_utc > 0) {
        pk->time_offset_valid = 1;
        pk->time_offset = network_utc - system_utc;
    }
    else {
        pk->time_offset_valid = pk->time_offset = 0;
    }
}

void pk_clear(struct upload_pack *pk)
{
    *pk->str = '\0';
    pk->used = 0;
}

void pk_free(struct upload_pack *pk)
{
    free(pk->str);
    pk->used = pk->ready = 0;
}

void pk_append(struct upload_pack *pk, char *str)
{
    /* if we're about to use more than we have, grow buffer */
    if (pk->used + strlen(str) >= pk->ready) {
        pk->ready += BUFSIZE;
        pk->str = realloc(pk->str, pk->ready);
    }

    strcpy(pk->str + pk->used, str);
    pk->used += strlen(str);
}

/**
 * Upload probing results to CMAND servers.
 *
 * @param pk The upload_pack to upload to CMAND
 * @return 1 if successful, -1 if not
 */
int pk_upload(struct upload_pack *pk)
{
    char *buf = NULL;
    char *post_skel = "POST /%s%s HTTP/1.1\r\n"
                      "Host: %s\r\n"
                      "User-Agent: %s\r\n"
                      "Content-type: application/x-www-form-urlencoded\r\n"
                      "Content-length: %d\r\n"
                      "\r\n"
                      "%s";
    int sock, err, maxbuf, ret_val = -1;
   
    /* allocate some extra to fit content-length param */
    maxbuf = strlen(post_skel) + strlen(CMAND_PREFIX) + strlen(CMAND_PATH_REPORT) + 
             strlen(CMAND_URL) + strlen(PACKAGE_STRING) + 15 + pk->used;
    buf = malloc(maxbuf);
    if (buf == NULL) fatal("Failed to allocate memory for data upload buffer!");
    snprintf(buf, maxbuf, post_skel, CMAND_PREFIX, CMAND_PATH_REPORT, 
             CMAND_URL, PACKAGE_STRING, pk->used, pk->str);

    sock = connect_to_cmand();
    if (sock < 0) goto cleanbuf;

    /* send the POST request */
    debug(LOW, "Sending %d bytes of results data via HTTP POST to %s...\n", pk->used, CMAND_URL);
    err = send(sock, buf, strlen(buf), 0);
    if (err < 0) goto cleanall;
    
    /* read the response */
    memset(buf, 0, maxbuf);
    err = recv_tmo(sock, buf, maxbuf, CMAND_TMO);
    if (err <= 0) goto cleanall;

    ret_val = 1;    /* currently we're happy so long as we get something back */
cleanall:
    net_close(sock);
cleanbuf:
    free(buf);
    return ret_val;
}

char *get_time_str(struct upload_pack *pk)
{
    static char t[25];
    time_t tm;

    tm = time(NULL);
    tm = tm + pk->time_offset;
    strftime(t, 25, "%Y-%m-%dT%H:%M:%S", gmtime(&tm));

    if (! pk->time_offset_valid) strcat(t, "*");
    return t;
}

