/********************************************************************
 * @file    hc.c
 * @brief   Front-end client to hiccups library
 * @author  $LastChangedBy: rcraven $ (hiccups@cmand.org)
 * @version $LastChangedRevision: 540 $
 * @date    $Date: 2014-07-28 22:25:21 -0700 (Mon, 28 Jul 2014) $
 *
 * $Id: hc.c 540 2014-07-29 05:25:21Z rcraven $
 ********************************************************************/
#include "hiccups.h"
#include "settings.h"

enum debugLevel verbose = DEF_VERBOSITY;
int want_privacy = 0;
int batch_mode = 0;
int plab_mode = 0;
FILE *resfile = NULL;

/**
 * Print usage help statement.
 *
 * @param Program name as in argv[0]
 */
static void usage(char *prog)
{
    printf("---------------------------------------------------\n");
    printf("| TCP HICCUPS Raw Socket Client - version %-8s|\n", VERSION);
    printf("| Written by: CMAND, hiccups@cmand.org            |\n");
    printf("---------------------------------------------------\n");
    printf("Usage: %s -q\n", prog);
    printf("       %s [Modes ...] [-i iface] [-s src_addr] [-t targ_file | target]\n", prog);
    printf("\n");
    printf("    -q           Query/interview user for settings (overrides all else)\n");
    printf("\n");
    printf("Modes:\n");
    printf("    -e           Enable ECN negotiation\n");
    printf("    -m           Enable MSS negotiation\n");
    printf("    -l val       MSS value (only use with -m)\n");
    printf("    -n           Enable WINSCL negotiation\n");
    printf("    -k           Enable SACK negotiation\n");
    printf("    -u           Enable TIMESTAMP negotiation\n");
    printf("    -y           Enable Multipath TCP negotiation\n");
    printf("    -x           Enable experimental TCP option\n");
    printf("    -a           Enable AppSalt mode\n");
    printf("    -d file      Application-layer data to send (only use with -a)\n");
    printf("    -c cvr       Specify a single field coverage type (%d for all)\n", HNUMPROBES);
    printf("    -v           Verbose mode (use twice for extra noisy)\n");
    printf("    -p           Privacy mode (do not report results to cmand.org)\n");
    printf("    -b           Batch mode (no DNS queries or version checks)\n");
    printf("    -w file      Append probing results to file\n");
    printf("    -z           PlanetLab mode (use only when querying a PL server)\n");
    printf("\n");
    printf("Source / Target Specifiers:\n");
    printf("    -i iface     Network interface to use for pcap (ignored on Linux)\n");
    printf("    -s addr      Source IPv4 address for probe (will guess if not provided)\n");
    printf("    -t           Treat last arg as path to newline-separated list of targets\n");
    printf(" [target[:port]] DNS/IPv4 address and TCP port to probe\n");
    printf("                 (uses TCP port %d by default)\n", DEF_TARG_PORT);
    printf("                 If omitted, download and probe list of targets from cmand.org\n");
    printf("\n");
    exit(EXIT_SUCCESS);
}

static int get_user_response(void)
{
    char buf[BUFSIZE];
    int buflen = BUFSIZE;

    if (get_line(buf, buflen) > 0) {
        if (strncasecmp(buf, "y", 1) == 0) return 1;
        if (strncasecmp(buf, "n", 1) == 0) return -1;
        if (strncasecmp(buf, "d", 1) == 0) return -2;
    }
    return 0;
}

static void reset_sk(struct hiccups_sock *sk)
{
    memset(sk, 0, sizeof (struct hiccups_sock));
    sk->local.sin_family = AF_INET;
    sk->local.sin_port = htons(DEF_SRC_PORT);
    sk->cvr = 99;
    sk->mssval = DEF_MSS_VAL;
    sk->appbuf = DEF_APP_MSG;
    sk->applen = strlen(DEF_APP_MSG);
    sk->sock = INVALID_SOCKET;
}

/**
 * Interview the user for settings.
 *
 * @param targlist The start of the target list
 * @param sk Holds settings for the hiccups library to use
 * @return The number of targets we initialized (always 1 for this mode)
 */
static int interview_mode(struct sockaddr_in **targlist, struct hiccups_sock *sk)
{
    char buf[BUFSIZE];
    int buflen = BUFSIZE;
    struct sockaddr_in tmp;

    printf("---------------------------------------------------\n");
    printf("| TCP HICCUPS Raw Socket Client - version %-8s|\n", VERSION);
    printf("| Written by: CMAND, hiccups@cmand.org            |\n");
    printf("---------------------------------------------------\n");
    color_printf(TERM_GREEN, "Interview Mode");
    printf(" - hint: leave line blank to use [] defaults\n");

#ifndef _WIN
    ensure_root_priv();
#endif
    reset_sk(sk);
    want_privacy = 2;   /* 2 means we're not sure yet; ask later */

    printf("Target address [%s] > ", DEF_TARG_ADDR);
    if (get_line(buf, buflen) == 0) strcpy(buf, DEF_TARG_ADDR);
    set_single_targ(buf, DEF_TARG_PORT, targlist);

    printf("Target port [%d] > ", DEF_TARG_PORT);
    if (get_line(buf, buflen) > 0) (*targlist)->sin_port = htons(validate_port(atoi(buf)));

    infer_my_ip(*targlist, &tmp);
    printf("Source address [%s] > ", inet_ntoa(tmp.sin_addr));
    if (get_line(buf, buflen) == 0) strcpy(buf, inet_ntoa(tmp.sin_addr));
    inet_aton(buf, &sk->local.sin_addr);

#ifndef _LINUX
    char *ifname;
    ifname = get_ifname(&sk->local);
    printf("Listen device for pcap [%s] > ", ifname);
    if (get_line(buf, buflen) == 0) sk->dev = ifname;
    else                            sk->dev = strdup(buf);
#endif

    printf("Will probe: %s:%d --> %s:%d\n", buf, ntohs(sk->local.sin_port),
        inet_ntoa((*targlist)->sin_addr), ntohs((*targlist)->sin_port));
    printf("Set advanced settings? [no] > ");
    if (get_user_response() != 1) return 1;
 
    printf("Negotiate ECN?         [no] > ");
    if (get_user_response() == 1) sk->ecn = 1;

    printf("Negotiate MSS?         [no] > ");
    if (get_user_response() == 1) sk->mss = 1;

    if (sk->mss) {
        printf("|_ MSS value         [%4d] > ", DEF_MSS_VAL);
        if (get_line(buf, buflen) > 0)
            sk->mssval = validate_port(atoi(buf));
    }

    printf("Use AppSalt?           [no] > ");
    if (get_user_response() == 1) sk->app = 1;

    if (sk->app) {
        printf("|_ Path to data file        > ");
        if (get_line(buf, buflen) > 0)
            sk->applen = read_data_file(buf, &(sk->appbuf));
    }

    printf("Coverage type?       [auto] > ");
    if (get_line(buf, buflen) > 0) sk->cvr = (char) atoi(buf);

    printf("Verbosity level (0 = OFF, 1 = LOW, 2 = HIGH) [0] > ");
    if (get_line(buf, buflen) == 0) strcpy(buf, "0");
    sscanf(buf, "%d", (int *) &verbose);

    return 1;
}

/**
 * Handle command line arguments.
 *
 * @param argc Passed from main()
 * @param argv Passed from main()
 * @param targlist Where to store the targets we load
 * @param sk Holds settings for the hiccups library to use
 * @return The number of targets we initialized
 */
static int process_args(int argc, char **argv, struct sockaddr_in **targlist, struct hiccups_sock *sk)
{
    int opt, num_targs, use_addr_file = 0;
    uint16_t tport = DEF_TARG_PORT;
    char *colon = NULL;

    /* process options-style (-X arg) arguments */
    while ((opt = getopt(argc, argv, "qeml:nkuyxad:c:vpbw:zi:s:th")) != -1) {
        switch ((char) opt) {
        case 'q':
            return interview_mode(targlist, sk);
            break;
        case 'e':
            sk->ecn = 1;
            break;
        case 'm':
            sk->mss = 1;
            break;
        case 'l':
            sk->mssval = validate_port(atoi(optarg));
            break;
        case 'n':
            sk->winscl = 1;
            break;
        case 'k':
            sk->sack = 1;
            break;
        case 'u':
            sk->tstamp = 1;
            break;
        case 'y':
            sk->mpath = 1;
            break;
        case 'x':
            sk->exopt = 1;
            break;
        case 'a':
            sk->app = 1;
            break;
        case 'd':
            sk->applen = read_data_file(optarg, &(sk->appbuf));
            break;
        case 'c':
            sk->cvr = (char) atoi(optarg);
            break;
        case 'v':
            verbose++;
            break;
        case 'p':
            want_privacy = 1;
            break;
        case 'b':
            batch_mode = 1;
            break;
        case 'w':
            resfile = fopen(optarg, "a");
            if (resfile == NULL) fatal("Could not open file %s for writing", optarg);
            break;
        case 'z':
            plab_mode = 1;
            break;
        case 'i':
            sk->dev = strdup(optarg);
            break;
        case 's':
            inet_aton(optarg, &sk->local.sin_addr);
            break;
        case 't':
            use_addr_file = 1;
            break;
    	case '?':
            if ((optopt == 'l') || (optopt == 'd') || (optopt == 'c') || 
                    (optopt == 'w') || (optopt == 'i') || (optopt == 's'))
                printf("Option -%c requires an argument.\n\n", optopt);
            else if (isprint (optopt))
                printf("Unknown option `-%c'.\n\n", optopt);
            else
                printf("Unknown option character `\\x%x'.\n\n", optopt);
            /* fall through */
        case 'h':
        default:
            usage(argv[0]);
        }
    }

#ifndef _WIN
    ensure_root_priv();
#endif

    /* setup our probing target(s) in 1 of 3 possible ways... */
    /* 1. Download a list from cmand.org */
    if (optind >= argc) {
        num_targs = set_targ_from_cmand(targlist);
    }
    /* 2. If -t was set, read a list in from a local file */
    else if (use_addr_file) {
        num_targs = set_targ_from_file(argv[optind], targlist);
    }
    /* 3. Use the non-option argument as the single target */
    else {
        colon = strrchr(argv[optind], ':');
        if (colon != NULL) {
            tport = validate_port(atoi(colon + 1));
            *colon = '\0';
        }
        num_targs = set_single_targ(argv[optind], tport, targlist);
    }
    if (! num_targs) fatal("Failed to set any targets!");

    /* infer our public-facing address if not given in args */
    if (sk->local.sin_addr.s_addr == 0)
        infer_my_ip(*targlist, &sk->local);

    return num_targs;
}

static void print_result(struct hiccups_sock *sk, int can_trust)
{
    if (! can_trust) printf("** ");
    printf("Result for ");
    color_printf(TERM_CYAN, "%-8s", cvr_names[sk->cvr]);
    printf(" probe: ");
    switch (sk->rescode) {
    case BOTHMATCH:
        color_printf(TERM_GREEN, "MATCHED both paths");
        printf(" (S: %s | SA: %s)", sk->syn_str, sk->sa_str);
        break;
    case SAMATCH:
        color_printf(TERM_YELLOW, "SYN-ACK matched, but SYN did not");
        printf(" (S: %s | SA: %s)", sk->syn_str, sk->sa_str);
        break;
    case SANULLMATCH:
        color_printf(TERM_MAGENTA, "SYN integrity blocked, but SYN-ACK passed HNULL");
        printf(" (S: %s | SA: %s)", sk->syn_str, sk->sa_str);
        break;
    case SYNMATCH:
        color_printf(TERM_BROWN, "SYN matched, but SYN-ACK did not");
        printf(" (S: %s | SA: %s)", sk->syn_str, sk->sa_str);
        break;
    case NONEMATCH:
        color_printf(TERM_RED, "FAILED to match either path");
        printf(" (S: %s | SA: %s)", sk->syn_str, sk->sa_str);
        break;
    case -1:
        printf("Connection failed");
        break;
    default:
        printf("TIMEOUT");
        break;
    }
    printf("\n");
}

static void write_resfile_line(struct hiccups_sock *sk, struct upload_pack *pk, int can_trust)
{
    if (resfile == NULL) fatal("Tried to write result, but file was not open");

    /* Output file:
     * - Each line reprsents a separate SYN/SYN-ACK pair
     * - Line format: time, targ_ip, targ_port, src_ip, src_port, modes, \
     *     cvr, cvr_str, rescode, rescode_str, syn_match, sa_match, can_trust
     *   - Modes: A=appsalt, E=ecn, S=sack_perm, T=timestamp, W=winscl, 
     *            P=MP_CAPABLE, X=opt99, M=mss (always followed by mssval as rest of field)
     * - Append an asterisk to the time if we couldn't get the network time
     */
    fprintf(resfile, "%s,", get_time_str(pk));
    fprintf(resfile, "%s,%d,", inet_ntoa(sk->remote.sin_addr), ntohs(sk->remote.sin_port));
    fprintf(resfile, "%s,%d,", inet_ntoa(sk->local.sin_addr), ntohs(sk->local.sin_port));
    fprintf(resfile, "%s,%d,%s,", get_mode_str(sk), sk->cvr, cvr_names[sk->cvr]);
    fprintf(resfile, "%d,%s,", sk->rescode, get_rescode_name(sk->rescode));
    fprintf(resfile, "%s,%s,%d\n",sk->syn_str, sk->sa_str, can_trust);
}

static void gen_post_string(struct hiccups_sock *sk, struct upload_pack *pk, int can_trust, char *buf)
{
    char *targ_ip = strdup(inet_ntoa(sk->remote.sin_addr));
    
    sprintf(buf, "time[]=%s&targ_ip[]=%s&targ_port[]=%d&src_ip[]=%s&src_port[]=%d" \
           "&modes[]=%s&cvr[]=%d&cvr_str[]=%s&res[]=%d&res_str[]=%s&syn_match[]=%s" \
           "&sa_match[]=%s&can_trust[]=%d&",
           get_time_str(pk), targ_ip, ntohs(sk->remote.sin_port),
           inet_ntoa(sk->local.sin_addr), ntohs(sk->local.sin_port), get_mode_str(sk),
           sk->cvr, cvr_names[sk->cvr], sk->rescode, get_rescode_name(sk->rescode),
           sk->syn_str, sk->sa_str, can_trust);

    free(targ_ip);
}

static void handle_results(struct hiccups_sock *sk, struct upload_pack *pk, int can_trust)
{
    char post_str[BUFSIZE];

    print_result(sk, can_trust);
    if (resfile != NULL) write_resfile_line(sk, pk, can_trust);

    if (want_privacy != 1) {
        gen_post_string(sk, pk, can_trust, post_str);
        pk_append(pk, post_str);
    }
}

int main(int argc, char **argv)
{
    struct hiccups_sock sk;
    struct upload_pack pk;
    struct sockaddr_in *targlist = NULL;
    int i, j, val, num_targ, can_trust;
    uint16_t tmp_port;

#ifdef _WIN
    init_winsock();
#endif

    reset_sk(&sk);
    num_targ = process_args(argc, argv, &targlist, &sk);
    if (! batch_mode) check_for_new_version();
    if (want_privacy != 1) pk_init(&pk);

    /* sanity checks */
    if ((targlist == NULL) || (num_targ <= 0))
        fatal("Invalid target");
    if (targlist[0].sin_addr.s_addr == 0)
        fatal("Invalid target");
    if (sk.local.sin_addr.s_addr == 0)
        fatal("Invalid source IP");

    /* perform the requested probe(s) */
    for (i = 0; i < num_targ; i++) {
        sk.remote = *(targlist + i);  /* set the target ip:port */

        if (num_targ > 1) printf("Probing target %s:%d...\n",
            inet_ntoa(sk.remote.sin_addr), ntohs(sk.remote.sin_port));

        /* Option 1: User wants a single probe using the specified cvr */
        if ((sk.cvr >= 0) && (sk.cvr < HNUMPROBES)) {
            hiccups_probe(&sk);
            handle_results(&sk, &pk, 1);
        }
        /* Option 2: User wants all probes done sequentially (debug mode) */
        else if (sk.cvr == HNUMPROBES) {
            /* Try a null probe first so we know if to trust results */
            sk.cvr = HNULL;
            hiccups_probe(&sk);
            if (! sk.samatch) can_trust = 0;
            else              can_trust = 1;
            handle_results(&sk, &pk, can_trust);

            tmp_port = ntohs(sk.local.sin_port) + 1;
            for (j=0; j < HNUMPROBES-1; j++) {
                sk.cvr = j;
                sk.local.sin_port = htons(tmp_port + j); /* increment src_port between each probe */
                hiccups_probe(&sk);
                handle_results(&sk, &pk, can_trust);
            }
            sk.cvr = HNUMPROBES; /* reset for next target */
            sk.local.sin_port = htons(tmp_port - 1);
            if (! can_trust) printf("** HICCUPS won't work on this path; results likely invalid!\n");
        }
        /* Option 3: Auto probe */
        else {
            hiccups_auto_probe(&sk);
        }

        /* sleep and upload between each target */
        if (i < num_targ-1) {
            sleep(1);
            if (! want_privacy) {
                pk_upload(&pk);
                pk_clear(&pk);
            }
        }
    }

    /* if we did interview mode, we won't assume anything yet */
    if (want_privacy == 2) {
        printf("Would you like to contribute these results to cmand.org? Y/n/d(etails first) > ");
        if ((val = get_user_response()) >= 0) want_privacy = 0;
        if (val == -2) {    /* details first */
            printf("\nONLY the following data will be transmitted: (%d bytes)\n%s\n",
                   pk.used, pk.str);
            printf("\nProceed with upload? [yes] > ");
            if (get_user_response() >= 0) want_privacy = 0;
        }
    }

    if (! want_privacy) {
        val = pk_upload(&pk);
        if (val)    printf("Reporting successful.  Thank you for your contribution!\n");
        else        printf("Failed to upload results: %s\n", strerror(errno));
    }

    if (want_privacy != 1) pk_free(&pk);
    if (resfile != NULL) fclose(resfile);
    exit(EXIT_SUCCESS);
}

