/********************************************************************
 * @file    hiccups.c
 * @brief   HICCUPS hashing routines (should mirror kernel patch)
 * @author  $LastChangedBy: rcraven $ (hiccups@cmand.org)
 * @version $LastChangedRevision: 535 $
 * @date    $Date: 2014-06-30 17:12:51 -0700 (Mon, 30 Jun 2014) $
 *
 * $Id: hiccups.c 535 2014-07-01 00:12:51Z rcraven $
 ********************************************************************/
#include "hiccups.h"
#include "settings.h"

/**
 * Populates an hb with a specific TCP option.
 *
 * @param ip The packet's IPv4 header
 * @param tcp The packet's TCP header
 * @param hb Destination to where the options should be copied
 * @param kind The kind number of the TCP option to copy
 * @return Length of hb to end of options
 */
static int populate_hb_tcpopt(struct ip *ip, struct tcphdr *tcp, struct hiccups_bits *hb, uint8_t kind)
{
    unsigned char *dest, *byte;
    int full_len, opt_len, copied, pos;

    copied = pos = 0;
    dest = (unsigned char *) &(hb->options);
    byte = ((unsigned char *) tcp) + 20;
    full_len = (tcp->th_off << 2) - 20;

    /* scan the TCP options and copy any with the specified kind byte */
    while (pos < full_len) {
        if (*byte == 0) break;
        opt_len = *(byte + 1);
        if (*byte != kind) {
            if (*byte == 1) opt_len = 1;
            pos += opt_len;
            byte += opt_len;
            continue;
        }
        memcpy(dest, byte, opt_len);
        pos += opt_len;
        copied += opt_len;
        byte += opt_len;
        dest += opt_len;
    }

    return (sizeof (*hb) - sizeof (hb->options)) + copied;
}

/**
 * Copies IP and TCP options to an hb.
 *
 * @param ip The packet's IPv4 header
 * @param tcp The packet's TCP header
 * @param hb Destination to where the options should be copied
 * @return Full length of hb after copying options
 */
static int populate_hb_options(struct ip *ip, struct tcphdr *tcp, struct hiccups_bits *hb)
{
    unsigned char *dest, *ip_options_start, *tcp_options_start;
    int ip_options_len, tcp_options_len;    /* in octets */
    
    /* locate the IP and TCP options */
    dest = (unsigned char *) &(hb->options);
    ip_options_start  = ((unsigned char *) ip) + 20;
    ip_options_len    = ((unsigned char *) tcp) - ip_options_start;
    tcp_options_start = ((unsigned char *) tcp) + 20;
    tcp_options_len   = (tcp->th_off << 2) - 20;

    /* load IP and TCP options into hiccups_bits struct */
    memcpy(dest, ip_options_start, ip_options_len);
    memcpy(dest + ip_options_len, tcp_options_start, tcp_options_len);

    /* calculate hiccups_bits length to end of options */
    return (sizeof (*hb) - sizeof (hb->options)) + ip_options_len + tcp_options_len;
}

/**
 * Populate IP addresses and TCP ports into an hb.
 *
 * @param ip IPv4 header of the packet to hash
 * @param tcp TCP header of the packet to hash
 * @param hb Hashing structure to populate
 */
static void populate_hb_addrport(struct ip *ip, struct tcphdr *tcp, struct hiccups_bits *hb)
{
    /* all fields in hb are in network order */
    hb->saddr = ip->ip_src.s_addr;
    hb->daddr = ip->ip_dst.s_addr;
    hb->sport = tcp->th_sport;
    hb->dport = tcp->th_dport;
}

/**
 * Populate length-related fields into an hb.
 * 
 * @param ip IPv4 header of the packet to hash
 * @param tcp TCP header of the packet to hash
 * @param hb Hashing structure to populate
 */
static void populate_hb_len(struct ip *ip, struct tcphdr *tcp, struct hiccups_bits *hb)
{
    hb->ihl = ip->ip_hl;
    hb->doff = tcp->th_off;
    hb->tot_len = ip->ip_len;
}

/**
 * Populate "safe" fields into an hb.  These are fields less likely to encounter changes.
 *
 * @param ip IPv4 header of the packet to hash
 * @param tcp TCP header of the packet to hash
 * @param hb Hashing structure to populate
 */
static void populate_hb_safe(struct ip *ip, struct tcphdr *tcp, struct hiccups_bits *hb)
{
    hb->version = ip->ip_v;
    hb->frag_off |= htons(IP_RF) & ip->ip_off;
    hb->protocol = ip->ip_p;
    hb->res1 |= 0xE & tcp->th_x2;
    hb->urg_ptr = tcp->th_urp;
}

/**
 * Populate ECN bits from the IP header into an hb.
 *
 * @param ip IPv4 header of the packet to hash
 * @param hb Hashing structure to populate
 */
static void populate_hb_ecnip(struct ip *ip, struct hiccups_bits *hb)
{
    uint8_t ecn_cp = IPTOS_ECN_MASK & ip->ip_tos;
    hb->ecn = ((ecn_cp & 0x02)>1) | (ecn_cp & 0x01);    /* only keep bitwise OR */
}

/**
 * Populates ECN bits from the TCP header into an hb.
 *
 * @param tcp TCP header of the packet to hash
 * @param hb Hashing structure to populate
 */
static void populate_hb_ecntcp(struct tcphdr *tcp, struct hiccups_bits *hb)
{
    hb->res1 |= 0x1 & tcp->th_x2;                       /* Nonce Sum bit */
    hb->flags |= 0xC0 & tcp->th_flags;                  /* CWR, ECE flags */
}

/**
 * Populates flags into an hb.
 *
 * @param ip IPv4 header of the packet to hash
 * @param tcp TCP header of the packet to hash
 * @param hb Hashing structure to populate
 */
static void populate_hb_flags(struct ip *ip, struct tcphdr *tcp, struct hiccups_bits *hb)
{
    hb->frag_off |= htons(IP_DF) & ip->ip_off;
    hb->flags |= 0x3F & tcp->th_flags;
}

/**
 * Populates all of the fixed-length (non-options) header fields into an hb.
 *
 * @param ip IPv4 header of the packet to hash
 * @param tcp TCP header of the packet to hash
 * @param hb Hashing structure to populate
 */
static void populate_hb_nonat_fixed(struct ip *ip, struct tcphdr *tcp, struct hiccups_bits *hb)
{
    populate_hb_safe(ip, tcp, hb);
    populate_hb_len(ip, tcp, hb);
    populate_hb_ecnip(ip, hb);
    populate_hb_ecntcp(tcp, hb);
    populate_hb_flags(ip, tcp, hb);
}

/**
 * Populates an hb for convenient hashing.
 *
 * @param ip IPv4 header of the packet to hash
 * @param tcp TCP header of the packet to hash
 * @param hb Hashing structure to populate
 * @param cvr Which fields to populate
 * @return Length of hb to end of options
 */
static int populate_hb(struct ip *ip, struct tcphdr *tcp, struct hiccups_bits *hb, uint8_t cvr)
{
    int len = sizeof (*hb) - sizeof (hb->options);   /* length of hb without any options */

    /* different things get populated depending on specified field coverage */
    switch (cvr) {
    case HFULL:
        populate_hb_addrport(ip, tcp, hb);
        populate_hb_nonat_fixed(ip, tcp, hb);
        len = populate_hb_options(ip, tcp, hb);
    case HNONAT:
        populate_hb_nonat_fixed(ip, tcp, hb);
        len = populate_hb_options(ip, tcp, hb);
        break;
    case HNAT:
        populate_hb_addrport(ip, tcp, hb);
        break;
    case HNOOPT:
        populate_hb_nonat_fixed(ip, tcp, hb);
        break;
    case HONLYOPT:
        len = populate_hb_options(ip, tcp, hb);
        break;
    case HECNIP:
        populate_hb_ecnip(ip, hb);
        break;
    case HECNTCP:
        populate_hb_ecntcp(tcp, hb);
        break;
    case HLEN:
        populate_hb_len(ip, tcp, hb);
        break;
    case HMSS:
        len = populate_hb_tcpopt(ip, tcp, hb, TCP_KIND_MSS);
        break;
    case HWINSCL:
        len = populate_hb_tcpopt(ip, tcp, hb, TCP_KIND_WINSCL);
        break;
    case HTSTAMP:
        len = populate_hb_tcpopt(ip, tcp, hb, TCP_KIND_TSTAMP);
        break;
    case HMPTCP:
        len = populate_hb_tcpopt(ip, tcp, hb, TCP_KIND_MPATH);
        break;
    case HEXOPT:
        len = populate_hb_tcpopt(ip, tcp, hb, TCP_KIND_EXP);
        break;
    case HFLAGS:
        populate_hb_flags(ip, tcp, hb);
        len = populate_hb_tcpopt(ip, tcp, hb, TCP_KIND_SACKPERM);
        break;
    case HSAFE:
        populate_hb_safe(ip, tcp, hb);
        break;
    case HNULL:
    case HONEWAY:
        break;
    default:
        fatal("Unknown field coverage type.");
    }

    hb->coverage = cvr;
    /* hb->salt = htons(ip->ip_id); */  /* old implementation */
    return len;
}

/**
 * Hashes selected fields from a packet's headers.
 *
 * @param ip IPv4 header of the packet to hash
 * @param tcp TCP header of the packet to hash
 * @param cvr Which fields to include in the hash
 * @param hash The resulting hiccups hashes
 */
static void hiccups_hash(struct ip *ip, struct tcphdr *tcp, uint8_t cvr, struct hiccups_hashes *hash)
{
    struct hiccups_bits hb;
    int len;

    memset(&hb, 0, sizeof (hb));
    len = populate_hb(ip, tcp, &hb, cvr);

    hash->crc = crc32(0, &hb, len);
    hash->murmur = murmur3(0, &hb, len);

    if (verbose >= HIGH) print_binary((unsigned char *) &hb, len, 4, 2);
    debug(HIGH, "\t\t[ cvr = %-10d  |  hb_len = %-10d ]\n", cvr, len);
    debug(HIGH, "\t\t[ crc = 0x%08x  |  murmur = 0x%08x ]\n", hash->crc, hash->murmur);
}

/**
 * Returns 1 if a majority of the lowest 8 bits are 1
 *
 * @param a The value to check
 * @return Whether the value has a majority of 1s
 */
static int majority_exists(int a)
{
    if ((a == 0x3) || (a == 0x5) || (a == 0x6) || (a == 0x7))
        return 1;
    else
        return 0;
}

/**
 * Compares two sets of hashes, returning how many matched
 *
 * @param h1 The first hash set to compare
 * @param h2 The second hash set to compare
 * @return Binary encoded number of matches
 */
static int compare_hashes(struct hash_pieces *h1, struct hash_pieces *h2)
{
    int match = 0;
    if ((h1->isn_upper & 0x0FFF) == (h2->isn_upper & 0x0FFF))
        match |= 0x4;
    if ((h1->ipid & 0x0FFF) == (h2->ipid & 0x0FFF))
        match |= 0x2;
    if ((h1->rcvwin & 0x0FFF) == (h2->rcvwin & 0x0FFF))
        match |= 0x1;
    return match;
}

/**
 * Determines most reasonable set of syn_match bits.
 *
 * @param sa_match The match bits computed by comparing hashes on the SYN-ACK
 * @param hp Copies of the integrity transmission fields from the SYN-ACK
 * @return The match bits the SYN receiver most likely computed
 */
static int get_syn_match(int sa_match, struct hash_pieces *hp)
{
    int isn_syn_match = 0xDD;
    int ipid_syn_match = 0xEE;
    int rcvwin_syn_match = 0xFF;

    /* In searching for a majority, we only consider syn_match candidates from 
     * integrity transmission fields marked in the SYN-ACK's match bits
     * -AND- have a 1 in the highest bit of the syn_match.
     */
    if ((sa_match & 0x4) && (hp->isn_upper & 0x8000))   /* isn was marked */
        isn_syn_match = (hp->isn_upper >> 12) & 0x7;
    if ((sa_match & 0x2) && (hp->ipid & 0x8000))        /* ipid */
        ipid_syn_match = (hp->ipid >> 12) & 0x7;
    if ((sa_match & 0x1) && (hp->rcvwin & 0x8000))      /* rcvwin */
        rcvwin_syn_match = (hp->rcvwin >> 12) & 0x7;

    /* now see if there's a majority amongst the valid candidates */
    /* if ((isn_syn_match == ipid_syn_match) && (isn_syn_match == rcvwin_syn_match)) return isn_syn_match; */
    if (isn_syn_match == ipid_syn_match) return isn_syn_match;
    if (isn_syn_match == rcvwin_syn_match) return isn_syn_match;
    if (ipid_syn_match == rcvwin_syn_match) return ipid_syn_match;
    /* else */
    return 0;
}

/**
 * Determines the appropriate value for each integrity transmission field.
 *
 * @param ip IPv4 header of the packet to hash
 * @param tcp TCP header of the packet to hash
 * @param cvr Which fields to include in the hash
 * @param hp Where to store the calculated hash values
 */
void calc_hashes(struct ip *ip, struct tcphdr *tcp, uint8_t cvr, struct hash_pieces *hp, uint8_t embed)
{
    struct hiccups_hashes hash;
    uint16_t shifted_cvr;

    hiccups_hash(ip, tcp, cvr, &hash);
    shifted_cvr = cvr << 12;
    if (embed) shifted_cvr = embed << 12;

    hp->isn_upper = shifted_cvr | (hash.crc & 0x00000FFF);
    hp->ipid      = shifted_cvr | (hash.murmur & 0x00000FFF);
    hp->rcvwin    = shifted_cvr | ((hash.murmur & 0xFFF00000) >> 20);

    /* protect the isn */
    hp->salt      = get_rand_bytes(2);
    hp->isn       = ((hp->salt ^ hp->isn_upper) << 16) | hp->salt;
}

static uint8_t find_hic_coverage(struct hiccups_sock *sk, struct ip *ip, struct tcphdr *tcp, struct hash_pieces *hp_rcvd)
{
    struct hash_pieces hp_calc;
    int i, cvr[3];
    uint8_t want_cvr, s_want_cvr = 0;
    uint8_t match, s_match = 0;
    uint8_t isn_cvr          = hp_rcvd->isn_upper >> 12;
    uint8_t ipid_cvr         = hp_rcvd->ipid >> 12;
    uint8_t rcvwin_cvr       = hp_rcvd->rcvwin >> 12;

    cvr[0] = isn_cvr;
    cvr[1] = ipid_cvr;
    cvr[2] = rcvwin_cvr;

    for (i = 0; i < 3; i++) {
        debug(HIGH, "\t|- Contents of hiccups_bits struct for cvr=%d:\n\t\t", cvr[i]);
        calc_hashes(ip, tcp, cvr[i], &hp_calc, 0);
        match = compare_hashes(hp_rcvd, &hp_calc);
        want_cvr = 0;
        if (match & 0x04) want_cvr = isn_cvr;
        if (match & 0x02) want_cvr = ipid_cvr;
        if (match & 0x01) want_cvr = rcvwin_cvr;

        /* only say we found the cvr used for the SYN hashes if at least 2 of 3 match */
        if (majority_exists(match)) {
            sk->synmatch = match;
            sk->srv_want_cvr = want_cvr;
            return cvr[i];
        }
        if (match > 0) {
            s_match = match;
            s_want_cvr = want_cvr;
        }
    }
    /* if no single hash matched for any of the checks, do majority
     * rules to try to infer what coverage they wanted */
    if (s_match == 0) {
        if (isn_cvr == ipid_cvr) s_want_cvr = isn_cvr;
        else if (isn_cvr == rcvwin_cvr) s_want_cvr = isn_cvr;
        else if (ipid_cvr == rcvwin_cvr) s_want_cvr = ipid_cvr;
        else s_want_cvr = HONEWAY; /* if no majority, send back HONEWAY */
    }
    sk->synmatch = s_match;
    sk->srv_want_cvr = s_want_cvr;
    return HONEWAY; /* there wasn't a match */
}

/**
 * Compare the received hashes with a hash of the SYN.
 *
 * @param sk Settings to use for this probe
 * @param pkt The packet buffer containing the SYN
 * @param len The true length of the SYN in the buffer
 */
void analyze_syn(struct hiccups_sock *sk, unsigned char *pkt, int len)
{
    struct ip *ip = (struct ip *) pkt;
    struct tcphdr *tcp;
    struct hash_pieces hp_rcvd;
    uint8_t cvr;

    /* ensure what we're looking at is a TCP SYN */
    if ((ip->ip_v != IPVERSION) || (ip->ip_p != IPPROTO_TCP)) return;
    tcp = (struct tcphdr *) (pkt + (ip->ip_hl << 2));
    if ((tcp->th_flags & TH_SYN) != TH_SYN) return;

    debug(LOW, "\n<< Received SYN << --------------------------------\n");

    /* break apart the SYN we just received */
    hp_rcvd.isn       = ntohl(tcp->th_seq);
    hp_rcvd.salt      = hp_rcvd.isn & 0x0000FFFF;
    hp_rcvd.isn_upper = (hp_rcvd.isn >> 16) ^ hp_rcvd.salt;
    hp_rcvd.ipid      = ntohs(ip->ip_id);
    hp_rcvd.rcvwin    = ntohs(tcp->th_win);

    /* print information about the packet */
    debug(LOW, "\t|- Source           : %s:%d\n", inet_ntoa(ip->ip_src), ntohs(tcp->th_sport));
    debug(LOW, "\t|- Destination      : %s:%d\n", inet_ntoa(ip->ip_dst), ntohs(tcp->th_dport));
    debug(LOW, "\t|- TCP seq (salted) : 0x%08x\n", hp_rcvd.isn);
    debug(LOW, "\t|- Upper ISN : 0x%04x | Salt   : 0x%04x\n", hp_rcvd.isn_upper, hp_rcvd.salt);
    debug(LOW, "\t|- IPID      : 0x%04x | RCVWIN : 0x%04x\n", hp_rcvd.ipid, hp_rcvd.rcvwin);
    debug(HIGH, "\n\t\t- Hex dump of SYN packet -\n");
    if (verbose >= HIGH) print_packet((const unsigned char *) ip, len);

    cvr = find_hic_coverage(sk, ip, tcp, &hp_rcvd);
    sk->cvr = cvr;
    if (cvr < HNUMPROBES) {
        int_to_binary(sk->synmatch, sk->syn_str);
        debug(LOW, "\t|- Result: HICCUPS match!\n");
        debug(LOW, "\t|-         Coverage: %s (%d)\n", cvr_names[cvr], cvr);
        debug(LOW, "\t|-         SYN match bits: %s\n", sk->syn_str);
        sk->rescode = SYNMATCH;
    }
    else {
        debug(LOW, "\t|- Result: Failed to match\n");
        sk->rescode = NONEMATCH;
    }
}

/**
 * Compare the received hashes with a hash of the SYN/ACK.
 *
 * @param sk Settings to use for this probe
 * @param pkt The packet buffer containing the SYN/ACK
 * @param len The true length of the SYN/ACK in the buffer
 */
void analyze_synack(struct hiccups_sock *sk, unsigned char *pkt, int len)
{
    struct ip *ip = (struct ip *) pkt;
    struct tcphdr *tcp;
    struct hash_pieces hp_rcvd, hp_calc;

    /* ensure what we're looking at is a SYN/ACK */
    if ((ip->ip_v != IPVERSION) || (ip->ip_p != IPPROTO_TCP)) return;
    tcp = (struct tcphdr *) (pkt + (ip->ip_hl << 2));
    if ((tcp->th_flags & (TH_SYN | TH_ACK)) != (TH_SYN | TH_ACK)) {
        sk->rescode = -1;   /* likely a RST: port closed */
        return;
    }

    /* break apart the SYN/ACK we just received */
    hp_rcvd.isn       = ntohl(tcp->th_seq);
    hp_rcvd.salt      = hp_rcvd.isn & 0x0000FFFF;
    hp_rcvd.isn_upper = (hp_rcvd.isn >> 16) ^ hp_rcvd.salt;
    hp_rcvd.ipid      = ntohs(ip->ip_id);
    hp_rcvd.rcvwin    = ntohs(tcp->th_win);

    /* print information about the packet */
    debug(LOW, "\t|- Source           : %s:%d\n", inet_ntoa(ip->ip_src), ntohs(tcp->th_sport));
    debug(LOW, "\t|- Destination      : %s:%d\n", inet_ntoa(ip->ip_dst), ntohs(tcp->th_dport));
    debug(LOW, "\t|- TCP seq (salted) : 0x%08x\n", hp_rcvd.isn);
    debug(LOW, "\t|- Upper ISN : 0x%04x | Salt   : 0x%04x\n", hp_rcvd.isn_upper, hp_rcvd.salt);
    debug(LOW, "\t|- IPID      : 0x%04x | RCVWIN : 0x%04x\n", hp_rcvd.ipid, hp_rcvd.rcvwin);
    debug(HIGH, "\n\t\t- Hex dump of SYN/ACK packet -\n");
    if (verbose >= HIGH) print_packet((const unsigned char *) ip, len);

    /* Final step: calculate hashes to check against the ones in the packet */

    /* If the forward path was unmodified, then cvr will be the same 
     * as was requested.  If the SYN was modified but the hiccups 
     * transmission fields were left intact, then cvr will also be 
     * the same as was requested.
     */
    debug(HIGH, "\n\t** Calculating hashes of SYN/ACK: *************\n");
    debug(HIGH, "\tTesting coverage type %s (%d):\n\t\t", cvr_names[sk->cvr], sk->cvr);
    calc_hashes(ip, tcp, sk->cvr, &hp_calc, 0);
    sk->samatch = compare_hashes(&hp_rcvd, &hp_calc);
    debug(HIGH, "\tFrom the calculated hash: (compared with received values above)\n");
    debug(HIGH, "\t|- Upper ISN : 0x%04x\n", hp_calc.isn_upper);
    debug(HIGH, "\t|- IPID      : 0x%04x\n", hp_calc.ipid);
    debug(HIGH, "\t|- RCVWIN    : 0x%04x\n", hp_calc.rcvwin);
    debug(HIGH, "\t|- Results of comparison: %d (the SYN-ACK match bits)\n", sk->samatch);
    if (majority_exists(sk->samatch)) {
        sk->synmatch = get_syn_match(sk->samatch, &hp_rcvd);
        debug(HIGH, "\t|- Majority found.  Extracted SYN_Match value: %d\n", sk->synmatch);
        if (majority_exists(sk->synmatch))
            sk->rescode = BOTHMATCH; /* both paths matched */
        else
            sk->rescode = SAMATCH; /* syn-ack matched but syn did not */
        return;
    }
    debug(HIGH, "\tDidn't match at least 2 of the 3 (ISN, IPID, RWIN).  Checking HONEWAY...\n");
    
    /* Or, if the SYN was completely trashed (meaning our hashes 
     * themselves were modified), then cvr will be HONEWAY (NULL).
     */
    debug(HIGH, "\tTesting coverage type %s (%d):\n\t\t", "HONEWAY", HONEWAY);
    calc_hashes(ip, tcp, HONEWAY, &hp_calc, 0);
    sk->samatch = compare_hashes(&hp_rcvd, &hp_calc);
    debug(HIGH, "\tFrom the calculated hash: (compared with received values above)\n");
    debug(HIGH, "\t|- Upper ISN : 0x%04x\n", hp_calc.isn_upper);
    debug(HIGH, "\t|- IPID      : 0x%04x\n", hp_calc.ipid);
    debug(HIGH, "\t|- RCVWIN    : 0x%04x\n", hp_calc.rcvwin);
    debug(HIGH, "\t|- Results of comparison: %d (the SYN-ACK match bits)\n", sk->samatch);
    if (majority_exists(sk->samatch)) {
        sk->synmatch = get_syn_match(sk->samatch, &hp_rcvd);
        debug(HIGH, "\t|- Majority found.  Extracted SYN_Match value: %d\n", sk->synmatch);
        sk->rescode = SANULLMATCH; /* odd case, syn was trashed (2 of 3 int. trans. fields modified) */
        /* syn-ack did't check anything but at least the syn-ack's int. trans. fields
         * weren't modified */
        return;
    }
    debug(HIGH, "\tDidn't match at least 2 of the 3 (ISN, IPID, RWIN).  SYN-ACK integrity failed.\n");

    /* Else, SYN-ACK integrity failed */
    sk->synmatch = get_syn_match(0x7, &hp_rcvd);
    debug(HIGH, "\t|- So, did the SYN at least match?  Extracted SYN_Match value: %d\n", sk->synmatch);
    if (majority_exists(sk->synmatch))
        sk->rescode = SYNMATCH; /* syn matched but syn-ack did not */
    else
        sk->rescode = NONEMATCH; /* both paths failed to match */
}

