/********************************************************************
 * @file    hiccups.h
 * @brief   User-space hiccups implementation
 * @author  $LastChangedBy: rcraven $ (hiccups@cmand.org)
 * @version $LastChangedRevision: 512 $
 * @date    $Date: 2014-05-27 00:14:19 -0700 (Tue, 27 May 2014) $
 *
 * $Id: hiccups.h 512 2014-05-27 07:14:19Z rcraven $
 ********************************************************************/

#if HAVE_CONFIG_H
  #include "config.h"
#endif

#ifndef _HICCUPS_H_
#define _HICCUPS_H_

/* All OSes use these */
#include <unistd.h>
#include <stdlib.h>

/* BSD, OSX, and Linux use these */
#ifndef _WIN

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <time.h>
#include <fcntl.h>
#include <signal.h>

#include <sys/types.h>
#include <strings.h>
#include <ctype.h>

#ifndef __FAVOR_BSD
    #define  __FAVOR_BSD
#endif

    #include <sys/socket.h>
    #include <netinet/ip.h>
    #include <netinet/tcp.h>
    #include <arpa/inet.h>
    #include <netdb.h>

    #include <netinet/in.h>
    #include <netinet/in_systm.h>
    #include <netinet/udp.h>
    #ifdef HAVE_NETINET_IP_VAR_H
        #include <netinet/ip_var.h>
    #endif
    #ifdef HAVE_NETINET_UDP_VAR_H
        #include <netinet/udp_var.h>
    #endif

    typedef int SOCKET;
    #define INVALID_SOCKET -1
    #define SOCKET_ERROR -1
#else /* _WIN */
    /* Windows uses these (ordering is important) */
    #define WPCAP /* to load winpcap-specific functions (like pcap_getevent()) */
    #define HAVE_REMOTE
    #include <winsock2.h>
    #include <windows.h>
    #include <wincrypt.h>
    #include <iphlpapi.h>
    #include <ws2tcpip.h>
    #include <dnet.h>

/* Windows doesn't have the TCP/IP header structs */
struct __attribute__ ((__packed__)) ip {
#ifdef HAVE_BIG_ENDIAN
    uint8_t ip_v:4;        /* version */
    uint8_t ip_hl:4;       /* header length */
#else
    uint8_t ip_hl:4;       /* header length */
    uint8_t ip_v:4;        /* version */
#endif
    uint8_t ip_tos;        /* type of service */
    uint16_t ip_len;       /* total length */
    uint16_t ip_id;        /* identification */
    uint16_t ip_off;       /* fragment offset field */
#define IP_RF 0x8000       /* reserved fragment flag */
#define IP_DF 0x4000       /* dont fragment flag */
#define IP_MF 0x2000       /* more fragments flag */
#define IP_OFFMASK 0x1fff  /* mask for fragmenting bits */
    uint8_t ip_ttl;        /* time to live */
    uint8_t ip_p;          /* protocol */
    uint16_t ip_sum;       /* checksum */
    struct in_addr ip_src, ip_dst;  /* source and dest address */
  };

struct __attribute__ ((__packed__)) tcphdr {
    uint16_t th_sport;     /* source port */
    uint16_t th_dport;     /* destination port */
    uint32_t th_seq;       /* sequence number */
    uint32_t th_ack;       /* acknowledgement number */
#ifdef HAVE_BIG_ENDIAN
    uint8_t th_off:4;      /* data offset */
    uint8_t th_x2:4;       /* (unused) */
#else
    uint8_t th_x2:4;       /* (unused) */
    uint8_t th_off:4;      /* data offset */
#endif
    uint8_t th_flags;
#define TH_FIN    0x01
#define TH_SYN    0x02
#define TH_RST    0x04
#define TH_PUSH   0x08
#define TH_ACK    0x10
#define TH_URG    0x20
    uint16_t th_win;       /* window */
    uint16_t th_sum;       /* checksum */
    uint16_t th_urp;       /* urgent pointer */
};
#define IPVERSION   4

#endif /* _WIN */

#ifndef _LINUX
    #include <pcap.h>
#else
    typedef int pcap_t;
#endif

#define func() do {if (DEBUG) fprintf(stdout,"\t>> %s:%s():%d\n",__FILE__,__FUNCTION__,__LINE__); } while (0)
#define warn(x...) do {fprintf(stderr,"*** Warn: "); fprintf(stderr,x); fprintf(stderr,"\n");} while (0)
#define fatal(x...) do {fprintf(stderr,"*** Fatal: "); fprintf(stderr,x); fprintf(stderr,"\n"); exit(-1);} while (0)
#define debug(level,x...) do {if (verbose >= level) {printf(x);} } while (0)

#define MAXPKT  80
#define BUFSIZE 1024    /**< Packet storage buffer, in octets. */
#define SNAPLEN 100     /**< Must be enough to cap all headers. */
#define TIMEOUT 2       /**< Listen timeout, in seconds. */

enum debugLevel {OFF, LOW, HIGH};
extern enum debugLevel verbose;    /* verbosity of hiccups library */
extern int batch_mode;
extern const char *cvr_names[];

/* HICCUPS socket data structure */
struct hiccups_sock {
    struct sockaddr_in local;   /* local address and port to use  */
    struct sockaddr_in remote;  /* remote address and port to use */
    uint32_t ecn:1,             /* negotiate ECN?             */
             mss:1,             /* negotiate MSS?             */
             winscl:1,          /* negotiate WINSCL?          */
             sack:1,            /* negotiate SACK?            */
             tstamp:1,          /* negotiate TSTAMP?          */
             mpath:1,           /* negotiate MPCAPABLE?       */
             exopt:1,           /* negotiate EXOPT?           */
             app:1,             /* use AppSalt?               */
             cvr:8,             /* coverage type              */
             mssval:16;         /* MSS value to request       */
    char *appbuf;               /* data to send if AppSalt    */
    int applen;                 /* length of data in appbuf   */
    int sock;                   /* raw socket descriptor      */
    char *dev;                  /* interface name for pcap    */
    pcap_t *pcap;               /* pcap interface             */
    int rescode;                /* HICCUPS result code        */
    int synmatch;               /* HICCUPS SYN match bits     */
    int samatch;                /* HICCUPS SYN-ACK match bits */
    char syn_str[4];            /* string form of synmatch    */
    char sa_str[4];             /* string form of samatch     */
    int srv_want_cvr;           /* server's want_coverage     */
};

struct upload_pack {
    char *str;              /* buffer to upload */
    int ready;              /* number of bytes allocated to str */
    int used;               /* number of bytes used by str so far */
    int time_offset_valid;  /* whether the system time's different from cmand */
    time_t time_offset;     /* how many seconds faster is cmand time */
};

/* probe types */
#define HNONAT   0
#define HFULL    1
#define HNAT     2
#define HNOOPT   3
#define HONLYOPT 4
#define HECNIP   5
#define HECNTCP  6
#define HLEN     7
#define HMSS     8
#define HWINSCL  9
#define HTSTAMP  10
#define HMPTCP   11
#define HEXOPT   12
#define HFLAGS   13
#define HSAFE    14
#define HNULL    15
#define HONEWAY  248
#define HNUMPROBES 16       /* don't count HONEWAY */
#define CVR_NAMES  {"HNONAT", "HFULL", "HNAT", "HNOOPT", "HONLYOPT", \
    "HECNIP", "HECNTCP", "HLEN", "HMSS", "HWINSCL", "HTSTAMP", "HMPTCP", \
    "HEXOPT", "HFLAGS", "HSAFE", "HNULL"}

/* result types */
#define NOTYETCHECK 0
#define BOTHMATCH   1
#define SAMATCH     2
#define SANULLMATCH 3
#define SYNMATCH    4
#define NONEMATCH   5
#define RES_NAMES   {"NOTYETCHECK", "BOTHMATCH", "SAMATCH", \
    "SANULLMATCH", "SYNMATCH", "NONEMATCH"}

#ifndef IPTOS_ECN_MASK
    #define IPTOS_ECN_MASK 0x03
#endif

/* terminal colors */
#ifndef _WIN
    /* these must be in the same order as in the #define that follows */
    enum COLORS {
        TERM_RED, TERM_YELLOW, TERM_GREEN, TERM_BLUE, TERM_BLACK, 
        TERM_CYAN, TERM_BROWN, TERM_MAGENTA, TERM_WHITE, TERM_NONE
    };
    /* 0; -> normal, 1; -> bold, 4; -> underline, 9; -> strike */
    #define ANSI_COLOR_STRINGS { \
    /* red */     "\033[1;31m", \
    /* yellow */  "\033[1;33m", \
    /* green */   "\033[1;32m", \
    /* blue */    "\033[1;34m", \
    /* black */   "\033[0;30m", \
    /* cyan */    "\033[1;36m", \
    /* brown */   "\033[0;34m", \
    /* magenta */ "\033[1;35m", \
    /* white */   "\033[1m", \
    /* none */    "\033[0m" }    /* flushes the previous property */
#else
    /* Windows pre-defines RGB, the rest is up to us */
    enum COLORS {
        TERM_BLACK        = 0,
        TERM_GREY         = FOREGROUND_BLUE+FOREGROUND_GREEN+FOREGROUND_RED,
        TERM_WHITE        = FOREGROUND_INTENSITY+TERM_GREY,
        TERM_BLUE         = FOREGROUND_INTENSITY+FOREGROUND_BLUE,
        TERM_GREEN        = FOREGROUND_INTENSITY+FOREGROUND_GREEN,
        TERM_RED          = FOREGROUND_INTENSITY+FOREGROUND_RED,
        TERM_CYAN         = FOREGROUND_INTENSITY+FOREGROUND_BLUE+FOREGROUND_GREEN,
        TERM_MAGENTA      = FOREGROUND_INTENSITY+FOREGROUND_RED+FOREGROUND_BLUE,
        TERM_BROWN        = FOREGROUND_RED+FOREGROUND_GREEN,
        TERM_YELLOW       = FOREGROUND_INTENSITY+TERM_BROWN,
        TERM_NONE         = TERM_GREY
    };
#endif 

/** TCP pseudo header for checksum. */
struct pseudo_header {
    uint32_t        src;
    uint32_t        dst;
    uint8_t         tmp;
    uint8_t         proto;
    uint16_t        tcp_len;
    uint8_t         tcp[60];
};

/* TCP options types */
#define TCP_KIND_NOP        1
#define TCP_KIND_MSS        2
#define TCP_KIND_WINSCL     3
#define TCP_KIND_SACKPERM   4
#define TCP_KIND_TSTAMP     8
#define TCP_KIND_MPATH      30
#define TCP_KIND_EXP        99

/** Holds the TCP Maximum Segment Size option. */
struct tcp_option_mss {
    uint8_t     kind;
    uint8_t     len;
    uint16_t    mss;
} __attribute__((packed));

/** Holds the TCP Receiver Window Scaling option. */
struct tcp_option_winscl {
    uint8_t     kind;
    uint8_t     len;
    uint8_t     scale;
} __attribute__((packed));

/** Holds the TCP SACK_Permitted option. */
struct tcp_option_sackperm {
    uint8_t     kind;
    uint8_t     len;
} __attribute__((packed));

/** Holds the TCP Timestamps option. */
struct tcp_option_tstamp {
    uint8_t     kind;
    uint8_t     len;
    uint32_t    tsval;
    uint32_t    tsecr;
} __attribute__((packed));

/** Holds the TCP Multipath Capable option. */
struct tcp_option_mpcapable {
    uint8_t     kind;
    uint8_t     len;
    uint8_t     version:4,
                subtype:4;
    uint8_t     flags;
    uint64_t    key;
} __attribute__((packed));

/** Holds data for hashing.  Same as tcp_hiccups.h in kernel. */
struct hiccups_bits {
    uint8_t     ihl:4,
                version:4;
    uint8_t     ecn:2,
                dscp:6;
    uint16_t    tot_len;
    uint16_t    frag_off;
    uint8_t     protocol;
    uint8_t     coverage;
    uint32_t    saddr;
    uint32_t    daddr;
    uint16_t    sport;
    uint16_t    dport;
    uint16_t    res1:4,
                doff:4,
                flags:8;
    uint16_t    window;
    uint16_t    urg_ptr;
    uint16_t    salt;
    /**
     * "options" includes enough room for 40 bytes each of IP and TCP options.
     * It must come last, because we won't always hash the whole thing.
     */
    uint8_t     options[80];
};

/** Holds the different types of hashes we do. */
struct hiccups_hashes {
    uint32_t murmur;
    uint32_t crc;
};

/** Holds hash pieces after being calculated. */
struct hash_pieces {
    uint32_t isn;
    uint16_t salt;
    uint16_t isn_upper;
    uint16_t ipid;
    uint16_t rcvwin;
};

/* publics from cmand.c */
void check_for_new_version(void);
int download_servlist(char **targlist);
void pk_init(struct upload_pack *pk);
void pk_clear(struct upload_pack *pk);
void pk_free(struct upload_pack *pk);
void pk_append(struct upload_pack *pk, char *str);
int pk_upload(struct upload_pack *pk);
char *get_time_str(struct upload_pack *pk);

/* publics from crc32.c */
uint32_t crc32(uint32_t crc, const void *buf, size_t size);

/* publics from hiccups.c */
void calc_hashes(struct ip *ip, struct tcphdr *tcp, uint8_t cvr, struct hash_pieces *hp, uint8_t embed);
void analyze_synack(struct hiccups_sock *sk, unsigned char *pkt, int len);

/* publics from missing.c */
#ifndef HAVE_STRSEP
    char* strsep(char **stringp, const char *delim);
#endif
#ifndef HAVE_INET_ATON
    int inet_aton(const char *ipstr, struct in_addr *out);
#endif
/*
#if HAVE_MALLOC == 0
    void *rpl_malloc(size_t n);
#endif
#if HAVE_REALLOC == 0
    void *rpl_realloc(void *p, size_t n);
#endif
*/

/* publics from murmur3.c */
uint32_t murmur3(uint32_t seed, const void *key, int len);

/* publics from net.c */
int raw_sock(struct sockaddr_in *sin_orig);
int infer_my_ip(const struct sockaddr_in *dst, struct sockaddr_in *myip);
void resolve_to_ip(char *url, struct sockaddr_in *target);
unsigned short in_cksum(unsigned short *addr, int len);
int recv_tmo(int sock, void *buf, size_t len, int timeout);
void net_close(int sock);
uint16_t validate_port(int port);
#ifndef _LINUX
    char *get_ifname(struct sockaddr_in *src);
#endif

/* publics from os_[linux|bsd|win].c */
#ifndef _LINUX
    void init_pcap(struct hiccups_sock *sk);
#endif
int listen_synack(struct hiccups_sock *sk, unsigned char **pkt);
int send_pkt(struct hiccups_sock *sk, unsigned char *pkt, int len);
#ifdef _WIN
    void init_winsock(void);
#endif

/* publics from packet.c */
uint32_t send_syn(struct hiccups_sock *sk);
void send_synack(struct hiccups_sock *sk, unsigned char *syn);
void send_rst(struct hiccups_sock *sk, uint32_t seqno, int is_ack);

/* publics from probe.c */
const char *get_rescode_name(int res);
char *get_mode_str(struct hiccups_sock *sk);
void hiccups_probe(struct hiccups_sock *sk);
void hiccups_auto_probe(struct hiccups_sock *sk);

/* publics from targlist.c */
int set_targ_from_cmand(struct sockaddr_in **target);
int set_targ_from_file(char *path, struct sockaddr_in **target);
int set_single_targ(char *addr, uint16_t port, struct sockaddr_in **target);

/* publics from util.c */
void print_binary(const unsigned char *buf, int len, int brk, int tabs);
void int_to_binary(int num, char *str);
void print_packet(const unsigned char *packet, int tot_len);
void color_printf(int color, const char *format, ...);
char *trim(char *s);
uint64_t get_rand_bytes(int bytes);
void ensure_root_priv(void);
int read_data_file(char *path, char **dest);
int get_line(char *buf, size_t len);

#endif  /* _HICCUPS_H_ */

