/********************************************************************
 * @file    hs.c
 * @brief   User-space raw socket HICCUPS server
 * @author  $LastChangedBy: rcraven $ (hiccups@cmand.org)
 * @version $LastChangedRevision: 535 $
 * @date    $Date: 2014-06-30 17:12:51 -0700 (Mon, 30 Jun 2014) $
 *
 * $Id: hs.c 535 2014-07-01 00:12:51Z rcraven $
 ********************************************************************/
#include "hiccups.h"
#include "settings.h"

enum debugLevel verbose = OFF;
int batch_mode = 0;
int quiet_mode = 0;
volatile sig_atomic_t runloop = 1;

static void sig_handler(int signum)
{
    runloop = 0;
}

/**
 * Print usage help statement.
 *
 * @param Program name (as in argv[0])
 */
static void usage(char *prog)
{
    printf("Server for HICCUPS Raw Socket Client\n");
    printf("---------------------------------------------\n");
    printf("Usage: %s [-v/vv] [-p <port>]\n", prog);
    printf("\n");
    printf("\t-p\tTCP port on which to listen (uses %d by default)\n", DEF_SERV_PORT);
    printf("\t-q\tQuiet mode (don't print anything)\n");
    printf("\t-v\tVerbose mode (use twice for extra noisy)\n");
    printf("\n");
    exit(EXIT_FAILURE);
}

/**
 * Handle command line arguments.
 *
 * @param argc Passed from main()
 * @param argv Passed from main()
 * @param The port to listen on
 */
static void process_args(int argc, char **argv, struct hiccups_sock *sk)
{
    int opt;

    /* process options-style (-X arg) arguments */
    while ((opt = getopt(argc, argv, "p:qvh")) != -1) {
        switch ((char) opt) {
        case 'p':
            sk->local.sin_port = htons(validate_port(atoi(optarg)));
            break;
        case 'q':
            quiet_mode++;
            break;
        case 'v':
            verbose++;
            break;
        case '?':
            printf("\n");
            /* fall through */
        case 'h':
        default:
            usage(argv[0]);
        }
    }
}

static void reset_sk(struct hiccups_sock *sk)
{
    memset(sk, 0, sizeof (struct hiccups_sock));
    sk->local.sin_family = sk->remote.sin_family = AF_INET;
    sk->local.sin_port = htons(DEF_SERV_PORT);
}

static void fw_rules(const struct hiccups_sock *sk, int install)
{
    char fwrule1[75], fwrule2[50];

    /* we can check which packets are ours via the magic TTL */
    snprintf(fwrule1, 75, FW1_MAGICTTL, MAGIC_TTL, ntohs(sk->local.sin_port));
    if (!install) fwrule1[10] = 'D';
    system(fwrule1);
    snprintf(fwrule2, 50, FW2_ELSEDROP, ntohs(sk->local.sin_port));
    if (!install) fwrule2[10] = 'D';
    system(fwrule2);
}

int main(int argc, char *argv[])
{
    struct hiccups_sock sk;
    int bytes_recv;
    struct ip *ip;
    struct tcphdr *tcp;
    unsigned char *pkt = NULL;

#ifndef _LINUX
    fatal("Sorry, the HICCUPS server currently only works on Linux");
#endif

    reset_sk(&sk);
    process_args(argc, argv, &sk);

    if (quiet_mode < 2) {
        printf("---------------------------------------------------\n");
        printf("| TCP HICCUPS Raw Socket Server - version %-8s|\n", VERSION);
        printf("| Written by: CMAND, hiccups@cmand.org            |\n");
        printf("---------------------------------------------------\n");
        printf("\n");
    }

    ensure_root_priv();
    //check_for_new_version();
    signal(SIGINT, sig_handler);

#ifndef _PLANETLAB
    /* block all packets from the kernel */
    fw_rules(&sk, 1);
    if (quiet_mode < 2) printf("iptables rules installed...\n");
    // TODO: if bsd, use ipfw
#endif

    pkt = (unsigned char *) malloc(1500);
    if (pkt == NULL) fatal("Failed to allocate memory for packet buffer!");

    sk.sock = raw_sock(&sk.local);

#ifdef _PLANETLAB
    /* HACKY AS HELL: we open a sock_stream and bind it to the same port so
     * that we can see packets with our PlanetLab raw socket.  The client
     * has to compensate by playing tricks with the TTL to differentiate
     * the two SYN-ACKs that will come back.
     *
     * I hate this, but have tried everything else I can think of and PL
     * staff are working on straightening this out for us, so this should
     * only be temporary (hopefully).
     */
    int wtf_sock = socket(AF_INET, SOCK_STREAM, 0);
    struct sockaddr_in sin = sk.local;
    sin.sin_addr.s_addr = htonl(INADDR_ANY);
    bind(wtf_sock, (struct sockaddr *) &sin, sizeof sin);
    listen(wtf_sock, 1);
#endif

    if (quiet_mode < 2) printf("listening on raw socket (port %d)...  (Press Ctrl+C to quit)\n", ntohs(sk.local.sin_port));

    while (runloop) {
        bytes_recv = recvfrom(sk.sock, pkt, 1500, 0, NULL, NULL);
        if (bytes_recv < 40) continue;

        ip = (struct ip *) pkt;
        tcp = (struct tcphdr *) (pkt + (ip->ip_hl << 2));
        if (tcp->th_dport != sk.local.sin_port) continue;
        if (tcp->th_flags & TH_RST) continue;

        sk.local.sin_addr = ip->ip_dst;
        sk.remote.sin_addr = ip->ip_src;
        sk.remote.sin_port = tcp->th_sport;

        /* if anything other than a pure SYN, emit a reset */
        if (tcp->th_flags != TH_SYN) {
            send_rst(&sk, tcp->th_seq, 1);
            continue;
        }

        /* OK, we've got a SYN packet we need to process */
        analyze_syn(&sk, pkt, bytes_recv);
        if (quiet_mode < 1) {
            printf("SYN probe from ");
            color_printf(TERM_CYAN, "%s:%d", inet_ntoa(sk.remote.sin_addr), ntohs(sk.remote.sin_port));
            printf(" -- ");
            if (sk.rescode == SYNMATCH) {
                color_printf(TERM_GREEN, "MATCHED");
                printf(" (%s)", sk.syn_str);
            }
            else {
                color_printf(TERM_RED, "NO MATCH");
            }
            printf("\n");
        }
        send_synack(&sk, pkt);
    }

    if (quiet_mode < 2) printf("\nCaught Ctrl+C.  Cleaning up...\n");
    free(pkt);
#ifndef _PLANETLAB
    fw_rules(&sk, 0);
    if (quiet_mode < 2) printf("iptables rules removed...\n");
#endif
    exit(EXIT_SUCCESS);
}

