/********************************************************************
 * @file    net.c
 * @brief   Network-related helper functions for hiccups
 * @author  $LastChangedBy: rcraven $ (hiccups@cmand.org)
 * @version $LastChangedRevision: 501 $
 * @date    $Date: 2014-05-01 00:28:44 -0700 (Thu, 01 May 2014) $
 *
 * $Id: net.c 501 2014-05-01 07:28:44Z rcraven $
 ********************************************************************/
#include "hiccups.h"

/**
 * Create a new raw socket.
 *
 * @param sin_orig Source port and address family (only used on PlanetLab)
 * @return Descriptor for the newly created socket (always invalid on Windows)
 */
int raw_sock(struct sockaddr_in *sin_orig)
{
#ifndef _WIN
    int sock, one = 1;

    /* establish raw socket */
    if ((sock = socket(PF_INET, SOCK_RAW, IPPROTO_TCP)) < 0)
        fatal("Create socket failed: %s", strerror(errno));

#ifdef _PLANETLAB
    /* create a new sin without the address */
    struct sockaddr_in sin = *sin_orig;
    sin.sin_addr.s_addr = htonl(INADDR_ANY);

    /* PlanetLab requires we bind to a port */
    if (bind(sock, (struct sockaddr *) &sin, sizeof sin) < 0)
        warn("bind failed: %s", strerror(errno));
#endif

    /* this lets us build our own IP headers */
    /* per PlanetLab raw socket API, this will only succeed after a successful bind */
    setsockopt(sock, 0, IP_HDRINCL, &one, sizeof one);

    return sock;

#else
    /* Raw sockets are pretty much useless on Windows.
     * Instead, we'll use libdnet to send and winpcap to receive. */
    return INVALID_SOCKET;
#endif /* ! _WIN */
}

/**
 * Determine what source address we should use for the given destination.
 *
 * @param dst The intented target address
 * @param myip Where to save the address we find
 * @return 1 for success, -1 for failure
 */
int infer_my_ip(const struct sockaddr_in *dst, struct sockaddr_in *myip)
{
    struct sockaddr_in me;
    socklen_t len = sizeof me;
    SOCKET sock;
    int err, ret = -1;

    sock = socket(AF_INET, SOCK_DGRAM, 0);
    if (sock == INVALID_SOCKET) return -1;

    err = connect(sock, (const struct sockaddr *) dst, sizeof (*dst));
    if (err == SOCKET_ERROR) goto clean;

    err = getsockname(sock, (struct sockaddr*) &me, &len);
    if (err == SOCKET_ERROR) goto clean;

    myip->sin_addr = me.sin_addr;
    ret = 1;

clean:
    net_close(sock);
    return ret;
}

/**
 * Lookup the given address in DNS.
 *
 * @param url DNS or IPv4 address in string form
 * @param target Where to save the IPv4 address we find
 */
void resolve_to_ip(char *url, struct sockaddr_in *target)
{
    int error;
    struct addrinfo hints, *result;

    /* hints allows us to tell getaddrinfo what kind of answer we want */
    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_INET;      /* we want IPv4 addresses only */

    /* we use getaddrinfo() because it can handle either DNS addresses or IPs as strings */
    error = getaddrinfo(url, NULL, &hints, &result);
    if (error) fatal("Error in getaddrinfo: %s", gai_strerror(error));

    /* just grab the first address in the linked list */
    target->sin_addr = ((struct sockaddr_in *) result->ai_addr)->sin_addr;

    freeaddrinfo(result);
}

/**
 * Compute an IP checksum.
 *
 * @param addr Start of the data range
 * @param len Length of the data range
 * @return 2-byte long IP checksum value
 */
unsigned short in_cksum(unsigned short *addr, int len)
{
    int             nleft = len;
    int             sum = 0;
    unsigned short  *w = addr;
    unsigned short  answer = 0;

    /*
     * Our algorithm is simple, using a 32 bit accumulator (sum), we add
     * sequential 16 bit words to it, and at the end, fold back all the
     * carry bits from the top 16 bits into the lower 16 bits.
     */
    while (nleft > 1) {
        sum += *w++;
        nleft -= 2;
    }

    /* 4mop up an odd byte, if necessary */
    if (nleft == 1) {
        *(unsigned char *)(&answer) = *(unsigned char *)w ;
        sum += answer;
    }

    /* 4add back carry outs from top 16 bits to low 16 bits */
    sum = (sum >> 16) + (sum & 0xffff); /* add hi 16 to low 16 */
    sum += (sum >> 16);         /* add carry */
    answer = ~sum;              /* truncate to 16 bits */
    return answer;
}

/**
 * Wraps recv with a timeout.
 *
 * @param sock Socket descriptor
 * @param buf Buffer to store the received data
 * @param len Length of the buffer
 * @param timeout Timeout value, in microseconds
 * @return int Number of bytes read, neg. on error
 */
int recv_tmo(int sock, void *buf, size_t len, int timeout)
{
    struct timeval t;
    int select_ret;
    fd_set socks;

    /* sock sets and timeval for select call */
    FD_ZERO(&socks);
    FD_SET(sock, &socks);
    t.tv_sec = 0;
    t.tv_usec = timeout;

    /* check to see if there's data we can read on the socket */
    select_ret = select(sock + 1, &socks, NULL, NULL, &t);

    /* bail if either there was an error or the timeout was reached */
    if (select_ret < 1) return select_ret;

    return recv(sock, buf, len, 0);
}

/**
 * Close a socket cross-platform.
 */
void net_close(int sock)
{
#ifndef _WIN
    close(sock);
#else
    closesocket(sock);
#endif
}

/**
 * Ensure the value is within the valid range.
 *
 * @param port
 * @return validated port number
 */
uint16_t validate_port(int port)
{
    if ((port < 1) || (port > 65535))
        fatal("Invalid number; must be between 1 and 65,535.");
    return (uint16_t) port;
}

#ifndef _LINUX
/**
 * Figures out the interface associated with an IP.
 *
 * @param src An IP address owned by one of the system's interfaces
 * @return The name of the associated interface (can be used with pcap)
 */
char *get_ifname(struct sockaddr_in *src)
{
    char errbuf[PCAP_ERRBUF_SIZE+1];
    char *ifname = NULL;
    pcap_if_t *alldevs;
    pcap_if_t *d;
    pcap_addr_t *a;

    /* retrieve the full list of interfaces */
    if (pcap_findalldevs(&alldevs, errbuf) == -1)
        fatal("pcap: pcap_findalldevs failed: %s\n", errbuf);

    /* scan the list for one with the requested IP */
    for (d = alldevs; d; d = d->next) {
        for (a = d->addresses; a; a = a->next) {
            if (a->addr->sa_family == AF_INET) {
                if (a->addr) {
                    if ( ((struct sockaddr_in *) a->addr)->sin_addr.s_addr == src->sin_addr.s_addr ) {
                        ifname = strdup(d->name);
                        goto done;
                    }
                }
            }
        }
    }
    fatal("pcap: No interface with addr %s was found.\n", inet_ntoa(src->sin_addr));

done:
    pcap_freealldevs(alldevs);
    return ifname;
}
#endif /* ! _LINUX */

