/********************************************************************
 * @file    os_bsd.c
 * @brief   BSD-specific functions
 * @author  $LastChangedBy: rcraven $ (hiccups@cmand.org)
 * @version $LastChangedRevision: 501 $
 * @date    $Date: 2014-05-01 00:28:44 -0700 (Thu, 01 May 2014) $
 *
 * $Id: os_bsd.c 501 2014-05-01 07:28:44Z rcraven $
 ********************************************************************/
#include "hiccups.h"

#ifdef _BSD

/**
 * Get the length of a data-link header.
 *
 * @param type The data-link type detected by pcap
 * @return Header length of the data-link type, in bytes 
 */
static int get_dlt_len(int type)
{
    switch (type) {
    case DLT_NULL:
        return 4;
        break;
    case DLT_RAW:
        return 0;
        break;
    case DLT_PPP:
        return 2;
        break;
    case DLT_EN10MB:
        return 14;
        break;
    default:
        warn("pcap: Unknown data-link type: %d\n", type);
        return 0;
        break;
    }
}

/**
 * Ensures that certain fields are in the appropriate byte order.
 *
 * @param ip The IP header to fix
 */
static void fix_byteorders(struct ip *ip)
{
    /* BSD is weird and needs these fields in HOST order to send
     * see: http://cseweb.ucsd.edu/~braghava/notes/freebsd-sockets.txt */
    ip->ip_len = ntohs(ip->ip_len);
    ip->ip_off = ntohs(ip->ip_off);
}

/**
 * Start a new pcap listener and engage the filter.
 *
 * @param sk pre-configured HICCUPS socket
 */
void init_pcap(struct hiccups_sock *sk)
{
    char errbuf[PCAP_ERRBUF_SIZE];
    char filter_exp[BUFSIZE];
    struct bpf_program fp;

    /* ensure we have a device picked out */
    if (sk->dev == NULL) fatal("pcap: No capture interface given");

    /* open the sniffer */
    errbuf[0] = '\0';
    sk->pcap = pcap_open_live(sk->dev, SNAPLEN, 0, 1, errbuf);
    /* IMPORTANT: we purposefully set a minimum read timeout here (1ms) because this timer has
     * a crappy behavior for what we're trying to do.  As an example, a 10s read timeout will
     * cause pcap_next/dispatch/loop to ALWAYS wait 10 seconds before returning--EVEN IF a
     * packet comes in right away.  This would slow down our probe tests considerably on BSD.
     *
     * Instead, we use pcap_get_selectable_fd(), select(), and fd_sets to manage the timeout
     * while still being able to return immediately once we get the first packet.
     */

    if (sk->pcap == NULL)
	    fatal("pcap_open_live failed: %s\n", errbuf);
    if (strlen(errbuf) > 0)
        debug(LOW, "pcap_open_live warning: %s\n", errbuf);

    /* filter for just our traffic */
    sprintf(filter_exp, "((tcp) and (dst port %d) and (src host %s))",
        ntohs(sk->local.sin_port), inet_ntoa(sk->remote.sin_addr));
    if (pcap_compile(sk->pcap, &fp, filter_exp, 1, 0) < 0)
        fatal("pcap_compile failed: %s\n", pcap_geterr(sk->pcap));
    if (pcap_setfilter(sk->pcap, &fp) < 0)
        fatal("pcap_setfilter failed: %s\n", pcap_geterr(sk->pcap));

    debug(LOW, "pcap: listening on interface: %s\n", sk->dev);
    debug(LOW, "pcap: using filter: %s\n", filter_exp);
}

/**
 * Use pcap to listen for the response from the probed host.
 * 
 * @param sk HICCUPS socket with pcap already initialized
 * @param pkt Buffer to hold the received packet
 * @return The length of the packet without any data link headers
 */
int listen_synack(struct hiccups_sock *sk, unsigned char **pkt)
{
    const unsigned char *frame = NULL;
    struct pcap_pkthdr hdr;
    struct timeval t;
    int ret, dlt_len, ippld_len, pcap_fd;
    fd_set pcaps;

    debug(LOW, "\n<< Receiving SYN/ACK << --------------------------------\n");

    dlt_len = get_dlt_len(pcap_datalink(sk->pcap));
    pcap_fd = pcap_get_selectable_fd(sk->pcap);

    do {
        /* reset fd set and timeout */
        FD_ZERO(&pcaps);
        FD_SET(pcap_fd, &pcaps);
        t.tv_sec = TIMEOUT;
        t.tv_usec = 0;

        /* look for a new packet we can read */
        ret = select(pcap_fd + 1, &pcaps, NULL, NULL, &t);
        if (ret < 0) fatal("** select() failed: %s", strerror(errno));

        /* either we found a packet to read or we timed out */
        if (ret) {
            frame = pcap_next(sk->pcap, &hdr);
            if (frame == NULL) fatal("pcap_next failed: %s\n", pcap_geterr(sk->pcap));
            debug(HIGH, "pcap: Sniffed packet. Pkt data: %p, caplen: %d, len: %d\n",
                frame, hdr.caplen, hdr.len);
            if (hdr.caplen >= dlt_len + 40) break; /* got a pkt with full headers */
            else
                debug(LOW, "pcap: Odd, IP payload was incomplete. DLT len: %d, caplen: %d\n",
                    dlt_len, hdr.caplen);
        }
        else {
            debug(LOW, "No reply.  %d sec timeout reached.\n", TIMEOUT);
            return 0;
        }
    } while (1);

    /* OK, we've successfully gotten the packet now */
    ippld_len = hdr.caplen - dlt_len;
    *pkt = malloc(ippld_len);
    memcpy(*pkt, frame + dlt_len, ippld_len);
    close(pcap_fd);
    pcap_close(sk->pcap);
    return ippld_len;
}

int send_pkt(struct hiccups_sock *sk, unsigned char *pkt, int len)
{
    /* BSD is weird and needs some fields in host order */
    fix_byteorders((struct ip *) pkt);

    return sendto(sk->sock, pkt, len, 0, (struct sockaddr *) &(sk->remote), sizeof (sk->remote));
}

#endif /* _BSD */

