/********************************************************************
 * @file    os_linux.c
 * @brief   Linux-specific functions
 * @author  $LastChangedBy: rcraven $ (hiccups@cmand.org)
 * @version $LastChangedRevision: 546 $
 * @date    $Date: 2014-10-07 08:10:28 -0700 (Tue, 07 Oct 2014) $
 *
 * $Id: os_linux.c 546 2014-10-07 15:10:28Z rcraven $
 ********************************************************************/
#include "hiccups.h"

#ifdef _LINUX

/**
 * Listen for the response from the probed host.
 *
 * @param sk Pre-configured HICCUPS socket
 * @param pkt Buffer to hold the received packet
 * @return The length of the packet without any data link headers
 */
int listen_synack(struct hiccups_sock *sk, unsigned char **pkt)
{
    struct ip *ip;
    struct tcphdr *tcp;
    struct timeval t;
    int bytes_recv, select_ret;
    fd_set socks;
    int plab_fake_syn_counter = 0;

    debug(LOW, "\n<< Receiving SYN/ACK << --------------------------------\n");

    /* search the receive buffer for a returning packet by matching the port */
    *pkt = malloc(BUFSIZE);
    do {
        /* sock sets and timeval for select call */
        FD_ZERO(&socks);
        FD_SET(sk->sock, &socks);
        t.tv_sec = TIMEOUT;
        t.tv_usec = 0;

        /* check to see if there's data we can read on the socket */
        if ((select_ret = select(sk->sock + 1, &socks, NULL, NULL, &t)) < 0)
            fatal("** select() failed: %s", strerror(errno));

        /* either the socket is ready or the timeout was reached */
        if (select_ret) {
            /* if the socket is ready, try to read from it */
            if ((bytes_recv = recvfrom(sk->sock, *pkt, BUFSIZE, 0, NULL, NULL)) < 0)
                fatal("** recvfrom() failed: %s", strerror(errno));
        }
        else {
            debug(LOW, "No reply.  %d sec timeout reached.\n", TIMEOUT);
            return 0;
        }

        /* make sure we got a full capture of the headers */
        if (bytes_recv < 40) continue;

        /* overlay header structs for while condition checks */
        ip = (struct ip *) *pkt;
        tcp = (struct tcphdr *) (*pkt + (ip->ip_hl << 2));

/*
 * RMC: This was an attempt to be able to probe an ./hs running on PlanetLab.
 *      but PlanetLab blocked everything I tried.  Commenting this out until we
 *      get more time to look at it.
 *
        /* HACKY HACK: PlanetLab mode *
        /* if synack and ttl less than 64, continue *
        if ((ip->ip_v == IPVERSION) && (ip->ip_p == IPPROTO_TCP)) { /* IPv4 & TCP packet *
            if ((tcp->th_flags & (TH_SYN | TH_ACK)) == (TH_SYN | TH_ACK)) { /* SYN-ACK *
		printf("i got a synack w/ ttl = %d\n", ip->ip_ttl);
                if (ip->ip_ttl < 75) {
                    if (plab_fake_syn_counter > 0) {
                        warn("in PlanetLab mode: I can't tell the difference between SYN/ACKs");
                        return 0;
                    }
                    plab_fake_syn_counter++;
                    tcp->th_dport = 0;
                    continue;
                }
            }
        }
*/
    } while (tcp->th_dport != sk->local.sin_port);

    return bytes_recv;
}

int send_pkt(struct hiccups_sock *sk, unsigned char *pkt, int len)
{
    return sendto(sk->sock, pkt, len, 0, (struct sockaddr *) &(sk->remote), sizeof (sk->remote));
}

#endif /* _LINUX */

