/********************************************************************
 * @file    os_win.c
 * @brief   MS Windows-specific functions
 * @author  $LastChangedBy: rcraven $ (hiccups@cmand.org)
 * @version $LastChangedRevision: 501 $
 * @date    $Date: 2014-05-01 00:28:44 -0700 (Thu, 01 May 2014) $
 *
 * $Id: os_win.c 501 2014-05-01 07:28:44Z rcraven $
 ********************************************************************/
#include "hiccups.h"

#ifdef _WIN

/**
 * Get the length of a data-link header.
 *
 * @param type The data-link type detected by pcap
 * @return Header length of the data-link type, in bytes 
 */
static int get_dlt_len(int type)
{
    switch (type) {
    case DLT_NULL:
        return 4;
        break;
    case DLT_RAW:
        return 0;
        break;
    case DLT_PPP:
        return 2;
        break;
    case DLT_EN10MB:
        return 14;
        break;
    default:
        warn("pcap: Unknown data-link type: %d\n", type);
        return 0;
        break;
    }
}

/**
 * Start a new pcap listener and engage the filter.
 *
 * @param sk Pre-configured HICCUPS socket
 */
void init_pcap(struct hiccups_sock *sk)
{
    char errbuf[PCAP_ERRBUF_SIZE];
    char filter_exp[BUFSIZE];
    struct bpf_program fp;

    /* ensure we have a device picked out */
    if (sk->dev == NULL) fatal("pcap: No capture interface given");

    /* open the sniffer */
    errbuf[0] = '\0';
    sk->pcap = pcap_open_live(sk->dev, SNAPLEN, 0, 1, errbuf);
    /* IMPORTANT: this timeout parameter in pcap_open_live is not the kind of timeout we
     * want, so we purposefully set this one to just 1ms.  This timeout controls how long
     * pcap should wait before processing a new incoming batch of packets.  As an example, 
     * a 10s timeout here will cause pcap_next/dispatch/loop to ALWAYS wait 10 seconds 
     * before returning--EVEN IF a packet comes in right away.  This would slow down our 
     * probe tests considerably.
     */

    if (sk->pcap == NULL)
	    fatal("pcap_open_live failed: %s\n", errbuf);
    if (strlen(errbuf) > 0)
        debug(LOW, "pcap_open_live warning: %s\n", errbuf);

    /* make Winpcap more responsive by lowering the event signal threshold to 40B from 16KB */
    pcap_setmintocopy(sk->pcap, 40);

    /* filter for just our traffic */
    sprintf(filter_exp, "((tcp) and (dst port %d) and (src host %s))",
        ntohs(sk->local.sin_port), inet_ntoa(sk->remote.sin_addr));
    if (pcap_compile(sk->pcap, &fp, filter_exp, 1, 0) < 0)
        fatal("pcap_compile failed: %s\n", pcap_geterr(sk->pcap));
    if (pcap_setfilter(sk->pcap, &fp) < 0)
        fatal("pcap_setfilter failed: %s\n", pcap_geterr(sk->pcap));

    debug(LOW, "pcap: listening on interface: %s\n", sk->dev);
    debug(LOW, "pcap: using filter: %s\n", filter_exp);
}

/**
 * Use pcap to listen for the response from the probed host.
 * 
 * @param sk HICCUPS socket with pcap already initialized
 * @param pkt Buffer to hold the received packet
 * @return The length of the packet without any data link headers
 */
int listen_synack(struct hiccups_sock *sk, unsigned char **pkt)
{
    /* A note about winpcap vs. libpcap:
     * Unfortunately, pcap_get_selectable_fd() is not available in winpcap.  There are also issues
     * with how Windows treats socket fds as compared to *nix (e.g. select() won't work on them).
     *
     * So, instead we use WaitForSingleObject() on the handle returned from pcap_getevent().
     * pcap_getevent() is a Winpcap-specific function.  A gotcha with it is that it won't
     * signal the handle unless 16KB of packet data has been received by the bpf, so we have
     * to lower this value to 40 bytes with pcap_setmintocopy(), which is also Winpcap-specific.
     */
    const unsigned char *frame = NULL;
    struct pcap_pkthdr hdr;
    int dlt_len, ippld_len;
    DWORD ret, msTimeout = TIMEOUT * 1000;

    debug(LOW, "\n<< Receiving SYN/ACK << --------------------------------\n");
    dlt_len = get_dlt_len(pcap_datalink(sk->pcap));

    do {
        /* look for a new packet we can read */
        ret = WaitForSingleObject(pcap_getevent(sk->pcap), msTimeout);
        if (ret == WAIT_FAILED) fatal("** wait() failed: error code %ld", GetLastError());

        /* either we found a packet to read or we timed out */
        if (ret == WAIT_OBJECT_0) {
            frame = pcap_next(sk->pcap, &hdr);
            if (frame == NULL) fatal("pcap_next failed: %s\n", pcap_geterr(sk->pcap));
            debug(HIGH, "pcap: Sniffed packet. Pkt data: %p, caplen: %d, len: %d\n",
                frame, hdr.caplen, hdr.len);
            if (hdr.caplen >= dlt_len + 40) break; /* got a pkt with full headers */
            else
                debug(LOW, "pcap: Odd, IP payload was incomplete. DLT len: %d, caplen: %d\n",
                    dlt_len, hdr.caplen);
        }
        else {
            debug(LOW, "No reply.  %d sec timeout reached.\n", TIMEOUT);
            return 0;
        }
    } while (1);

    /* OK, we've successfully gotten the packet now */
    ippld_len = hdr.caplen - dlt_len;
    *pkt = malloc(ippld_len);
    memcpy(*pkt, frame + dlt_len, ippld_len);
    pcap_close(sk->pcap);
    return ippld_len;
}

/**
 * Initialize the Winsock2 networking library.
 */
void init_winsock(void)
{
    WSADATA wsaData;
    int res;

    res = WSAStartup(MAKEWORD(2,2), &wsaData);
    if (res != 0) {
        printf("WSAStartup failed: %d\n", res);
        exit(-1);
    }
}

/**
 * Populates the intf_entry with details about the interface
 * that owns the provided IP.
 */
static void get_intf_of_ip(const struct sockaddr *ip, struct intf_entry *intf)
{
    const struct sockaddr_in *ip2 = (const struct sockaddr_in *) ip;
    struct addr sourceip;
    intf_t *i;

    /* convert source IP to dnet-style IP */
    addr_pton(inet_ntoa(ip2->sin_addr), &sourceip);

    i = intf_open();
    if (i == NULL) return;

    intf_get_src(i, intf, &sourceip);
}

/**
 * Uses libdnet to lookup the given IP in the OS's routing table.
 */
static int get_link_route(const struct sockaddr_in *dst, struct addr *gw)
{
    struct route_entry route;
    route_t *r;
    int is_gateway;

    r = route_open();
    if (r == NULL) return -1;

    /* load up the dst IP and check it with route_get */
    addr_pton(inet_ntoa(dst->sin_addr), &route.route_dst);
    if (route_get(r, &route) < 0) {
        *gw = route.route_dst;
        is_gateway = 0;
    }
    else {
        *gw = route.route_gw;
        is_gateway = 1;
    }

    route_close(r);
    return is_gateway;
}

/**
 * Uses libdnet to lookup the given IP in the OS's ARP table.
 * @return 1 = found it, 0 = not found, -1 = error
 */
static int lookup_in_arp_table(const struct addr *gw, struct addr *gwmac)
{
    struct arp_entry arp;
    struct addr zeromac;
    arp_t *a;
    int res;

    arp.arp_pa = *gw;
    addr_pton("00:00:00:00:00:00", &zeromac);

    a = arp_open();
    if (a == NULL) return -1;
    res = arp_get(a, &arp);
    arp_close(a);

    /* if we didn't find the MAC in the ARP table... */
    if ((res < 0) || (memcmp(arp.arp_ha.addr_data8, zeromac.addr_data8, ETH_ADDR_LEN) == 0))
        return 0;

    /* if we did, save it in gwmac */
    *gwmac = arp.arp_ha;
    return 1;
}

/**
 * Uses iphlpapi to tell Windows to send an ARP.
 */
static int arp_query(struct sockaddr_in *src, struct addr *gwip, struct addr *gwmac)
{
    int res, len = ETH_ADDR_LEN;
    uint32_t mac[2];

    res = SendARP(gwip->addr_ip, src->sin_addr.s_addr, (PULONG) mac, (PULONG) &len);
    if (res != NO_ERROR) return res;

    addr_pton("00:00:00:00:00:00", gwmac);
    memcpy(&(gwmac->addr_data8), (uint8_t *) mac, ETH_ADDR_LEN);
    return res;
}

/**
 * Windows replacement for raw sock sendto()
 * (we have to push an Ethernet frame through libdnet)
 */
int send_pkt(struct hiccups_sock *sk, unsigned char *pkt, int len)
{
    struct intf_entry intf;
    struct addr gwip, gwmac;
    unsigned char *frame = NULL;
    eth_t *e;

    /* find the interface that owns the source IP */
    get_intf_of_ip((const struct sockaddr *) &(sk->local), &intf);
    debug(HIGH, "dnet: %s is owned by interface %s (%s)\n", 
        inet_ntoa(sk->local.sin_addr), intf.intf_name, addr_ntoa(&(intf.intf_link_addr)));

    /* open a handle to write on that interface */
    e = eth_open(intf.intf_name);
    if (e == NULL) return -1;

    /* determine what link-local IP (gw) we send to for the given dst */
    get_link_route(&(sk->remote), &gwip);

    /* figure out the gw's MAC address */
    if (lookup_in_arp_table(&gwip, &gwmac) == 0) {
        arp_query(&(sk->local), &gwip, &gwmac);
    }
    debug(HIGH, "dnet: ARP'ed for local dest == %s\n", addr_ntoa(&gwmac));

    /* build the frame (Ethernet header + pkt) */
    frame = malloc(len + ETH_HDR_LEN);
    eth_pack_hdr(frame, gwmac.addr_eth, intf.intf_link_addr.addr_eth, ETH_TYPE_IP);
    memcpy(frame + ETH_HDR_LEN, pkt, len);

    /* send the full Ethernet frame */
    eth_send(e, frame, len + ETH_HDR_LEN);

    eth_close(e);
    free(frame);
    return 1;
}

#endif /* _WIN */

