/********************************************************************
 * @file    packet.c
 * @brief   Packet-crafting and listening functions
 * @author  $LastChangedBy: rcraven $ (hiccups@cmand.org)
 * @version $LastChangedRevision: 512 $
 * @date    $Date: 2014-05-27 00:14:19 -0700 (Tue, 27 May 2014) $
 *
 * $Id: packet.c 512 2014-05-27 07:14:19Z rcraven $
 ********************************************************************/
#include "hiccups.h"
#include "settings.h"

/**
 * Gets pointer to the IPv4 header of a packet.
 *
 * @param buf The packet buffer
 * @return The IPv4 headers
 */
static struct ip *get_iphdr(unsigned char *buf)
{
    return (struct ip *) buf;
}    

/**
 * Gets pointer to the TCP header of a packet.
 *
 * @param buf The packet buffer
 * @return The TCP headers
 */
static struct tcphdr *get_tcphdr(unsigned char *buf)
{
    struct ip *ip = get_iphdr(buf);

    /* depends on whether the packet has been built yet */
    if (ip->ip_hl > 0) return (struct tcphdr *) (buf + (ip->ip_hl << 2));
    else               return (struct tcphdr *) (buf + sizeof (struct ip));
}

/**
 * Gives a randomized ECN-enabled codepoint, 01 or 10.
 *
 * @return a 2 bit ECN codepoint in the lowest two bits
 */
static uint8_t rand_ecn_codepoint()
{
    if (get_rand_bytes(1) % 2)  return 0x01;
    else                        return 0x02;
}

/**
 * Builds the desired TCP option in the existing packet buffer.
 *
 * @param opt_start The original start of the TCP options space
 * @param len The current running length of the options, in bytes
 * @param kind The kind value of the TCP option to add
 * @param mssval Desired MSS value (only req'd when kind==MSS)
 * @return The new length of the options after appending
 */
static int append_tcp_option(unsigned char *opt_start, int len, uint8_t kind, uint16_t mssval)
{
    unsigned char *offset = opt_start + len;
    struct tcp_option_mss *tmss       = (struct tcp_option_mss *) offset;
    struct tcp_option_winscl *twinscl = (struct tcp_option_winscl *) offset;
    struct tcp_option_sackperm *tsack = (struct tcp_option_sackperm *) offset;
    struct tcp_option_tstamp *ttstamp = (struct tcp_option_tstamp *) offset;
    struct tcp_option_mpcapable *tmpc = (struct tcp_option_mpcapable *) offset;

    switch (kind) {
    case TCP_KIND_NOP:
        *offset = kind;
        len += 1;
        break;
    case TCP_KIND_MSS:
        tmss->kind = kind;
        tmss->len = 4;
        tmss->mss = htons(mssval);
        len += tmss->len;
        break;
    case TCP_KIND_WINSCL:
        twinscl->kind = kind;
        twinscl->len = 3;
        twinscl->scale = 7;
        len += twinscl->len;
        break;
    case TCP_KIND_SACKPERM:
        tsack->kind = kind;
        tsack->len = 2;
        len += tsack->len;
        break;
    case TCP_KIND_TSTAMP:
        ttstamp->kind = kind;
        ttstamp->len = 10;
        ttstamp->tsval = get_rand_bytes(4);
        ttstamp->tsecr = 0;
        len += ttstamp->len;
        break;
    case TCP_KIND_MPATH:
        tmpc->kind = kind;
        tmpc->len = 12;
        tmpc->subtype = 0;
        tmpc->version = 0;
        tmpc->flags = 0x81;
        tmpc->key = get_rand_bytes(8);
        len += tmpc->len;
        break;
    case TCP_KIND_EXP:
        tmss->kind = kind;
        tmss->len = 4;
        tmss->mss = get_rand_bytes(2);
        len += tmss->len;
        break;
    /* otherwise, do nothing */
    }
    return len;
}

/**
 * Ensures TCP options end on 4-byte boundary.
 *
 * @param opt_start The original start of the TCP options space
 * @param len The current running length of the options, in bytes
 * @return The new length of the options after filling with NOPs
 */
static int fill_tcp_options(unsigned char *opt_start, int len)
{
    unsigned char *offset = opt_start + len;
    int i, nops_needed;

    if (len % 4)
        nops_needed = 4 - (len % 4);
    else
        nops_needed = 0;

    for (i = 0; i < nops_needed; i++)
        *(offset + i) = 1;  /* NOP kind = 1 */

    return len + nops_needed;
}

/**
 * Calculates the IP and TCP checksums and stores in place
 *
 * @param ip The packet's IPv4 headers
 * @param tcp The packet's TCP headers
 */
static void fix_checksums(struct ip *ip, struct tcphdr *tcp)
{
    struct pseudo_header psh;
    unsigned short tlen;
    int plen;

    tlen = tcp->th_off << 2;
    plen = sizeof (psh) - sizeof (psh.tcp) + tlen;

    /* fill a pseudoheader for TCP checksum computation */
    psh.src = (ip->ip_src).s_addr;
    psh.dst = (ip->ip_dst).s_addr;
    psh.tmp = 0;
    psh.proto = ip->ip_p;
    psh.tcp_len = htons(tlen);
    memcpy(&psh.tcp, tcp, tlen);
    
    /* calculate IP and TCP checksums */
    ip->ip_sum = in_cksum((unsigned short *)ip, 20);
    tcp->th_sum = in_cksum((unsigned short *)&psh, plen);
}

/**
 * Populates a packet buffer with a (mostly) built SYN packet
 *
 * @param s Settings to use for this probe
 * @param buf The packet buffer to populate
 * @return The length of the built packet, in bytes
 */
static int build_syn(struct hiccups_sock *sk, unsigned char *buf)
{
    struct ip *ip                = get_iphdr(buf);
    struct tcphdr *tcp           = get_tcphdr(buf);
    unsigned char *tcp_opt_start = (unsigned char *) (tcp + 1);
    int optlen = 0;
    
    tcp->th_sport = sk->local.sin_port;
    tcp->th_dport = sk->remote.sin_port;
    tcp->th_off = sizeof (struct tcphdr) >> 2;
    tcp->th_flags = TH_SYN;

    if (sk->ecn) {
        ip->ip_tos = rand_ecn_codepoint();
        tcp->th_flags |= 0xC0;
    }

    /* append options like Linux does */
    if (sk->mss)     optlen = append_tcp_option(tcp_opt_start, optlen, TCP_KIND_MSS, sk->mssval);
    if (sk->sack && sk->tstamp) {
        optlen = append_tcp_option(tcp_opt_start, optlen, TCP_KIND_SACKPERM, 0);
        optlen = append_tcp_option(tcp_opt_start, optlen, TCP_KIND_TSTAMP, 0);
    }
    else if (sk->sack) {
        optlen = append_tcp_option(tcp_opt_start, optlen, TCP_KIND_NOP, 0);
        optlen = append_tcp_option(tcp_opt_start, optlen, TCP_KIND_NOP, 0);
        optlen = append_tcp_option(tcp_opt_start, optlen, TCP_KIND_SACKPERM, 0);
    }
    else if (sk->tstamp) {
        optlen = append_tcp_option(tcp_opt_start, optlen, TCP_KIND_NOP, 0);
        optlen = append_tcp_option(tcp_opt_start, optlen, TCP_KIND_NOP, 0);
        optlen = append_tcp_option(tcp_opt_start, optlen, TCP_KIND_TSTAMP, 0);
    }
    if (sk->winscl)  {
        optlen = append_tcp_option(tcp_opt_start, optlen, TCP_KIND_NOP, 0);
        optlen = append_tcp_option(tcp_opt_start, optlen, TCP_KIND_WINSCL, 0);
    }
    if (sk->mpath)   optlen = append_tcp_option(tcp_opt_start, optlen, TCP_KIND_MPATH, 0);
    if (sk->exopt)   optlen = append_tcp_option(tcp_opt_start, optlen, TCP_KIND_EXP, 0);
    /* optlen = fill_tcp_options(tcp_opt_start, optlen); */ /* don't need now */
    tcp->th_off += optlen >> 2;

    ip->ip_v = IPVERSION;
    ip->ip_hl = sizeof (struct ip) >> 2;
    ip->ip_len = htons(40 + optlen);
    ip->ip_off = htons(IP_DF);
    ip->ip_ttl = MAGIC_TTL;
    ip->ip_p = IPPROTO_TCP;
    ip->ip_src = sk->local.sin_addr;
    ip->ip_dst = sk->remote.sin_addr;

    return 40 + optlen;
}

/**
 * Craft and send a hiccups-enabled SYN packet.
 *
 * @param sk The already activated HICCUPS socket on which to send
 * @return the sequence number we used for this SYN
 */
uint32_t send_syn(struct hiccups_sock *sk)
{
    unsigned char *syn;
    struct ip *ip;
    struct tcphdr *tcp;
    struct hash_pieces hp;
    int len;

    syn = (unsigned char *) malloc(MAXPKT);
    if (syn == NULL) fatal("Failed to allocate memory for SYN buffer!");
    memset(syn, 0, MAXPKT);

    debug(LOW, "\n>> Sending HICCUPS SYN Probe >> ------------------------\n");

    len = build_syn(sk, syn);
    ip  = get_iphdr(syn);
    tcp = get_tcphdr(syn);

    /* determine hiccups fields */
    debug(LOW, "\t|- Source           : %s:%d\n", inet_ntoa(ip->ip_src), ntohs(tcp->th_sport));
    debug(LOW, "\t|- Destination      : %s:%d\n", inet_ntoa(ip->ip_dst), ntohs(tcp->th_dport));
    debug(HIGH, "\t|- Contents of hiccups_bits struct:\n\t\t");
    calc_hashes(ip, tcp, sk->cvr, &hp, 0);

    /* load the hiccups integrity transmission fields */
    tcp->th_seq = htonl(hp.isn);
    ip->ip_id   = htons(hp.ipid);
    tcp->th_win = htons(hp.rcvwin);

    fix_checksums(ip, tcp);

    /* print information about the packet */
    debug(LOW, "\t|- TCP seq (salted) : 0x%08x\n", hp.isn);
    debug(LOW, "\t|- Upper ISN : 0x%04x | Salt   : 0x%04x\n", hp.isn_upper, hp.salt);
    debug(LOW, "\t|- IPID      : 0x%04x | RCVWIN : 0x%04x\n", hp.ipid, hp.rcvwin);
    debug(HIGH, "\n\t\t- Hex dump of SYN packet -\n");
    if (verbose >= HIGH) print_packet(syn, len);

    /* send the SYN packet */
    if (send_pkt(sk, syn, len) < 0)
        fatal("** send_pkt() failed: %s", strerror(errno));
    free(syn);

    return hp.isn;
}

void send_synack(struct hiccups_sock *sk, unsigned char *syn)
{
    unsigned char *synack;
    struct ip *ip, *syn_ip;
    struct tcphdr *tcp, *syn_tcp;
    struct hash_pieces hp;
    int len = 40;
    
    debug(LOW, "\n>> Sending HICCUPS SYN-ACK Response >> ------------------------\n");

    /* setup: alloc synack and cast syn, synack */
    synack = (unsigned char *) malloc(MAXPKT);
    if (synack == NULL) fatal("Failed to allocate memory for SYN-ACK buffer!");
    memset(synack, 0, MAXPKT);
    ip  = get_iphdr(synack);
    tcp = get_tcphdr(synack);
    syn_ip  = get_iphdr(syn);
    syn_tcp = get_tcphdr(syn);

    /* build synack */
    tcp->th_sport = sk->local.sin_port;
    tcp->th_dport = sk->remote.sin_port;
    tcp->th_off   = sizeof (struct tcphdr) >> 2;
    tcp->th_flags = TH_SYN | TH_ACK;
    tcp->th_ack   = htonl(ntohl(syn_tcp->th_seq) + 1);

    ip->ip_v   = IPVERSION;
    ip->ip_hl  = sizeof (struct ip) >> 2;
    ip->ip_len = htons(len);
    ip->ip_off = htons(IP_DF);
    ip->ip_ttl = MAGIC_TTL;
    ip->ip_p   = IPPROTO_TCP;
    ip->ip_src = sk->local.sin_addr;
    ip->ip_dst = sk->remote.sin_addr;

    debug(LOW, "\t|- Source           : %s:%d\n", inet_ntoa(ip->ip_src), ntohs(tcp->th_sport));
    debug(LOW, "\t|- Destination      : %s:%d\n", inet_ntoa(ip->ip_dst), ntohs(tcp->th_dport));

    /* determine hiccups fields */
    debug(HIGH, "\t|- Contents of hiccups_bits struct:\n\t\t");
    calc_hashes(ip, tcp, sk->srv_want_cvr, &hp, sk->synmatch | 0x08);

    /* load the hiccups integrity transmission fields */
    tcp->th_seq = htonl(hp.isn);
    ip->ip_id   = htons(hp.ipid);
    tcp->th_win = htons(hp.rcvwin);

    fix_checksums(ip, tcp);

    /* print information about the packet */
    debug(LOW, "\t|- TCP seq (salted) : 0x%08x\n", hp.isn);
    debug(LOW, "\t|- Upper ISN : 0x%04x | Salt   : 0x%04x\n", hp.isn_upper, hp.salt);
    debug(LOW, "\t|- IPID      : 0x%04x | RCVWIN : 0x%04x\n", hp.ipid, hp.rcvwin);
    debug(HIGH, "\n\t\t- Hex dump of SYN-ACK packet -\n");
    if (verbose >= HIGH) print_packet(synack, len);

    /* send the SYN/ACK packet */
    if (send_pkt(sk, synack, len) < 0)
        fatal("** send_pkt() failed: %s", strerror(errno));
    free(synack);
}

void send_rst(struct hiccups_sock *sk, uint32_t seqno, int is_ack)
{
    unsigned char *rst;
    struct ip *ip;
    struct tcphdr *tcp;
    int len = 40;

    rst = (unsigned char *) malloc(len);
    if (rst == NULL) fatal("Failed to allocate memory for RST buffer!");
    memset(rst, 0, len);
    ip  = get_iphdr(rst);
    tcp = get_tcphdr(rst);

    tcp->th_sport = sk->local.sin_port;
    tcp->th_dport = sk->remote.sin_port;
    tcp->th_off   = sizeof (struct tcphdr) >> 2;
    if (is_ack) {
        tcp->th_flags = TH_RST | TH_ACK;
        tcp->th_ack = htonl(seqno + 1);
    }
    else {
        tcp->th_flags = TH_RST;
        tcp->th_seq = htonl(seqno + 1);
    }

    ip->ip_v   = IPVERSION;
    ip->ip_hl  = sizeof (struct ip) >> 2;
    ip->ip_len = htons(len);
    ip->ip_off = htons(IP_DF);
    ip->ip_id  = get_rand_bytes(2);
    ip->ip_ttl = MAGIC_TTL;
    ip->ip_p   = IPPROTO_TCP;
    ip->ip_src = sk->local.sin_addr;
    ip->ip_dst = sk->remote.sin_addr;

    fix_checksums(ip, tcp);

    if (send_pkt(sk, rst, len) < 0)
        fatal("** send_pkt() failed %s", strerror(errno));
    free(rst);
}

