/********************************************************************
 * @file    probe.c
 * @brief   HICCUPS client probing routines
 * @author  $LastChangedBy: rcraven $ (hiccups@cmand.org)
 * @version $LastChangedRevision: 502 $
 * @date    $Date: 2014-05-01 15:18:04 -0700 (Thu, 01 May 2014) $
 *
 * $Id: probe.c 502 2014-05-01 22:18:04Z rcraven $
 ********************************************************************/
#include "hiccups.h"
#include "settings.h"

const char *cvr_names[] = CVR_NAMES;
const char *res_names[] = RES_NAMES;
int mods_detected;

/**
 * Reset any special requests from the user.
 */
static void clear_modes(struct hiccups_sock *sk)
{
    sk->ecn = sk->mss = sk->winscl = sk->sack = sk->tstamp = sk->mpath = sk->exopt = 0;
}

/**
 * Reset any result code-related fields.
 */
static void clear_results(struct hiccups_sock *sk)
{
    sk->rescode = sk->synmatch = sk->samatch = 0;
}

/**
 * Wrapper for hiccups_probe to do it n times and then make sure it's the same.
 */
static void hiccups_probe_n(struct hiccups_sock *sk, int n)
{
    int last_rcode = -1;

    hiccups_probe(sk);
    if (sk->rescode < 0) return;

    if (last_rcode != sk->rescode)
        return;
    else
        last_rcode = sk->rescode;
}

const char *get_rescode_name(int res)
{
    if ((res > 0) && (res <= 5)) return res_names[res];
    else if (res == -1)          return "CLOSED";
    else                         return "TIMEOUT";
}

char *get_mode_str(struct hiccups_sock *sk)
{
    static char m[30];

    *m = '\0';
    if (sk->app)    strcat(m, "A");
    if (sk->ecn)    strcat(m, "E");
    if (sk->sack)   strcat(m, "S");
    if (sk->tstamp) strcat(m, "T");
    if (sk->winscl) strcat(m, "W");
    if (sk->mpath)  strcat(m, "P");
    if (sk->exopt)  strcat(m, "X");
    if (sk->mss)    snprintf(m + strlen(m), 13, "M%d", sk->mssval);

    return m;
}

/**
 * Perform a single probe.
 *
 * - Handles all socket setup and pcap (if needed)
 * - Sends special HICCUPS-enabled SYN packet to the target
 * - Looks for returning SYN-ACK and runs checks to determine path integrity
 *
 * @param sk Use settings from HICCUPS socket and save the results here
 */
void hiccups_probe(struct hiccups_sock *sk)
{
    unsigned char *synack = NULL;
    int retries, bytes_recv;
    uint32_t seqno;

    clear_results(sk);
#ifndef _WIN
    sk->sock = raw_sock(&(sk->local));  /* our caller must already ensure we have root */
#endif
#ifndef _LINUX
    if (sk->dev == NULL) sk->dev = get_ifname(&(sk->local));
    init_pcap(sk);
#endif

    retries = NUM_SYN_RETRIES;
    while (retries > 0) {
        seqno = send_syn(sk);
        bytes_recv = listen_synack(sk, &synack);
        if (bytes_recv > 0) {
            send_rst(sk, seqno, 0);
            analyze_synack(sk, synack, bytes_recv);
            break;
        }
        retries--;
        if (retries == 0) sk->rescode = -2;   /* likely a timeout */
    }
    
    int_to_binary(sk->synmatch, sk->syn_str);
    int_to_binary(sk->samatch, sk->sa_str);

#ifndef _WIN
    close(sk->sock);
#endif
    free(synack);
}

static void print_mod(const char *msg)
{
    mods_detected++;
    color_printf(TERM_RED, "MOD: ");
    printf(msg);
}

/**
 * Perform the minimum number of probes necessary to discover all mods.
 *
 * @param sk Use settings from HICCUPS socket and save the results here
 */
void hiccups_auto_probe(struct hiccups_sock *sk)
{
    char unidir[8];
    int nullcode, probes = 0;

    printf("---------------------------------------------------\n");
    printf("| TCP HICCUPS Raw Socket Client - version %-8s|\n", VERSION);
    printf("| Written by: CMAND, hiccups@cmand.org            |\n");
    printf("---------------------------------------------------\n");
    printf("\n");

    //clear_modes(sk);
    mods_detected = -1;

    /* Probe 1: we always do a NULL probe to check for HICCUPS ability */
    sk->cvr = HNULL;
    hiccups_probe(sk);
    probes++;
    nullcode = sk->rescode;

    color_printf(TERM_BLUE, "HICCUPS ability check: ");
    if (sk->rescode == -2) {
        color_printf(TERM_RED, "FAIL");
        printf(" -- Timeout.  Couldn't connect to server.\n");
        goto doneprobing;
    }
    if (sk->rescode == -1) {
        color_printf(TERM_RED, "FAIL");
        printf(" -- Connection refused.  Is this port open?\n");
        goto doneprobing;
    }
    if (sk->rescode == NONEMATCH) {
        color_printf(TERM_RED, "FAIL");
        printf(" -- HICCUPS won't work on this path!\n" \
            "\tThis could be due to either of the following two conditions:\n" \
            "\t - The other host is not HICCUPS-capable\n" \
            "\t - 2 or more of (ISN, IPID, RCVWIN) are being modified on either direction\n");
        goto doneprobing;
    }
    if (sk->rescode == SANULLMATCH) {
        color_printf(TERM_RED, "FAIL");
        printf(" -- This path is really weird.\n" \
        "\tAll of (ISN, IPID, RCVWIN) are being modified on the SYN, but the\n" \
        "\tSYN/ACK is fine.  Even though the SYN/ACK's HICCUPS fields are intact,\n" \
        "\tthe SYN fields are trashed so badly that it prevents us from getting\n" \
        "\tany reliable results.  Ultimately, HICCUPS won't work for this\n" \
        "\tbidirectional path.\n");
        goto doneprobing;
    }
    *unidir = '\0';
    if (sk->rescode == SYNMATCH) {
        color_printf(TERM_YELLOW, "HALF");
        printf(" -- HICCUPS only works on the SYN path.\n");
        strcpy(unidir, "SYN ");
    }
    if (sk->rescode == SAMATCH) {
        color_printf(TERM_YELLOW, "HALF");
        printf(" -- HICCUPS only works on the SYN/ACK path.\n");
        strcpy(unidir, "SYNACK ");
    }
    if (sk->rescode == BOTHMATCH) {
        color_printf(TERM_GREEN, "PASS");
        printf(" -- HICCUPS works well on this path!\n");
    }

    mods_detected = 0;
    color_printf(TERM_BLUE, "\n%sModifications Detected:\n", unidir);

    /* Check to see if any of the HICCUPS fields were modified */
    if (sk->synmatch == 6) print_mod("TCP Receive Window modified on SYN path\n");
    if (sk->synmatch == 5) print_mod("IPID modified on SYN path\n");
    if (sk->synmatch == 3) print_mod("TCP ISN modified on SYN path\n");
    if (sk->samatch == 6) print_mod("TCP Receive Window modified on SYN/ACK path\n");
    if (sk->samatch == 5) print_mod("IPID modified on SYN/ACK path\n");
    if (sk->samatch == 3) print_mod("TCP ISN modified on SYN/ACK path\n");

    /* Probe 2: Check everything (we assume that most paths have no mods) */
    sk->cvr = HFULL;
    hiccups_probe(sk);
    probes++;
    if (sk->rescode == nullcode) goto doneprobing;

    /* Probe 3: Check for NAT */
    sk->cvr = HNONAT;
    hiccups_probe(sk);
    probes++;
    if (sk->rescode == nullcode) {
        print_mod("Network Address Translation (NAT)\n");
        goto doneprobing;
    }

    /* Check IP/TCP options */
    sk->cvr = HONLYOPT;
    hiccups_probe(sk);
    probes++;
    if (sk->rescode == nullcode) goto checkfixedlen;
    /* is there a way we can avoid doing all of these once we find a winner? */
    sk->cvr = HMSS;
    hiccups_probe(sk);
    probes++;
    if (sk->rescode != nullcode) {
        if (sk->rescode == NONEMATCH) print_mod("MSS modified both ways\n");
        if (sk->rescode == SYNMATCH) print_mod("MSS modified on SYN/ACK\n");
        if (sk->rescode == SAMATCH) print_mod("MSS modified on SYN\n");
    }
    sk->cvr = HWINSCL;
    hiccups_probe(sk);
    probes++;
    if (sk->rescode != nullcode) {
        if (sk->rescode == NONEMATCH) print_mod("WINSCL modified both ways\n");
        if (sk->rescode == SYNMATCH) print_mod("WINSCL modified on SYN/ACK\n");
        if (sk->rescode == SAMATCH) print_mod("WINSCL modified on SYN\n");
    }
    sk->cvr = HTSTAMP;
    hiccups_probe(sk);
    probes++;
    if (sk->rescode != nullcode) {
        if (sk->rescode == NONEMATCH) print_mod("TIMESTAMP modified both ways\n");
        if (sk->rescode == SYNMATCH) print_mod("TIMESTAMP modified on SYN/ACK\n");
        if (sk->rescode == SAMATCH) print_mod("TIMESTAMP modified on SYN\n");
    }
    sk->cvr = HMPTCP;
    hiccups_probe(sk);
    probes++;
    if (sk->rescode != nullcode) {
        if (sk->rescode == NONEMATCH) print_mod("Multipath TCP modified both ways\n");
        if (sk->rescode == SYNMATCH) print_mod("Multipath TCP modified on SYN/ACK\n");
        if (sk->rescode == SAMATCH) print_mod("Multipath TCP modified on SYN\n");
    }
    sk->cvr = HEXOPT;
    hiccups_probe(sk);
    probes++;
    if (sk->rescode != nullcode) {
        if (sk->rescode == NONEMATCH) print_mod("EXOPT modified both ways\n");
        if (sk->rescode == SYNMATCH) print_mod("EXOPT modified on SYN/ACK\n");
        if (sk->rescode == SAMATCH) print_mod("EXOPT modified on SYN\n");
    }

    /* check to see if we need to check the fixed-lens individually */
    sk->cvr = HNOOPT;
    hiccups_probe(sk);
    probes++;
    if (sk->rescode == nullcode) goto doneprobing;

checkfixedlen:
    sk->cvr = HECNIP;
    hiccups_probe(sk);
    probes++;
    if (sk->rescode != nullcode) {
        if (sk->rescode == NONEMATCH) print_mod("ECN IP modified both ways\n");
        if (sk->rescode == SYNMATCH) print_mod("ECN IP modified on SYN/ACK\n");
        if (sk->rescode == SAMATCH) print_mod("ECN IP modified on SYN\n");
    }
    sk->cvr = HECNTCP;
    hiccups_probe(sk);
    probes++;
    if (sk->rescode != nullcode) {
        if (sk->rescode == NONEMATCH) print_mod("ECN TCP modified both ways\n");
        if (sk->rescode == SYNMATCH) print_mod("ECN TCP modified on SYN/ACK\n");
        if (sk->rescode == SAMATCH) print_mod("ECN TCP modified on SYN\n");
    }
    sk->cvr = HLEN;
    hiccups_probe(sk);
    probes++;
    if (sk->rescode != nullcode) {
        if (sk->rescode == NONEMATCH) print_mod("length modified both ways (an option was likely added or removed)\n");
        if (sk->rescode == SYNMATCH) print_mod("length modified on SYN/ACK (an option was likely added or removed)\n");
        if (sk->rescode == SAMATCH) print_mod("length modified on SYN (an option was likely added or removed)\n");
    }
    sk->cvr = HFLAGS;
    hiccups_probe(sk);
    probes++;
    if (sk->rescode != nullcode) {
        if (sk->rescode == NONEMATCH) print_mod("Flags modified both ways\n");
        if (sk->rescode == SYNMATCH) print_mod("Flags modified on SYN/ACK\n");
        if (sk->rescode == SAMATCH) print_mod("Flags modified on SYN\n");
    }
    sk->cvr = HSAFE;
    hiccups_probe(sk);
    probes++;
    if (sk->rescode != nullcode) {
        if (sk->rescode == NONEMATCH) print_mod("Reserved (or other scary) fields modified both ways\n");
        if (sk->rescode == SYNMATCH) print_mod("Reserved (or other scary) fields modified on SYN/ACK\n");
        if (sk->rescode == SAMATCH) print_mod("Reserved (or other scary)  fields modified on SYN\n");
    }

doneprobing:
    if (! mods_detected) color_printf(TERM_GREEN, "None detected.\n");
    printf("\n");
    printf("** Done probing **\n");
    printf("Number of probes used: %d\n", probes);
}

