/********************************************************************
 * @file    targlist.c
 * @brief   Functions to manage the list of target addrs
 * @author  $LastChangedBy: rcraven $ (hiccups@cmand.org)
 * @version $LastChangedRevision: 496 $
 * @date    $Date: 2014-04-17 19:23:40 -0700 (Thu, 17 Apr 2014) $
 *
 * $Id: targlist.c 496 2014-04-18 02:23:40Z rcraven $
 ********************************************************************/
#include "hiccups.h"
#include "settings.h"

/**
 * Parse a string in the form of ADDR:PORT into a target.
 *
 * @param str The string to parse
 * @param target The struct to load with the parsing results
 */
static void parse_str_to_targ(char *str, struct sockaddr_in *target)
{
    char *port_str;

    str = trim(str);
    target->sin_family = AF_INET;
    target->sin_port   = htons(DEF_TARG_PORT);

    /* if str contains a colon, we need to splice off the port */
    port_str = strrchr(str, ':');
    if (port_str != NULL) {
        target->sin_port = htons( validate_port( atoi(port_str + 1) ) );
        *port_str = '\0';
    }

    /* alright, now resolve the addr */
    resolve_to_ip(str, target);
}

/**
 * Set the list of targets to what's downloaded from cmand.org.
 *
 * @param target Pointer to the targets array to populate
 * @return The number of valid targets we set
 */
int set_targ_from_cmand(struct sockaddr_in **target)
{
    char *targ_list, *cur_targ, *next_targ;
    int num_targs, i = 0;

    num_targs = download_servlist(&targ_list);
    if (! num_targs) return 0;

    *target = calloc(num_targs, sizeof (struct sockaddr_in));

    /* tokenize the targ_list by line breaks */
    next_targ = targ_list;
    do { 
        cur_targ = strsep(&next_targ, "\r\n");
        if (*cur_targ == '\0') continue; /* cycle past empty strings */
        parse_str_to_targ(cur_targ, *target + i++);
    } while ((targ_list != NULL) && (i < num_targs));

    printf("** %d targets in the list downloaded from cmand.org\n", i);

    free(targ_list);
    return i;
}

/**
 * Load in probe targets from address input file.
 *
 * @param path The path to the address file that is to be read
 * @param target Pointer to the targets array to populate
 * @return The number of valid targets we set
 */
int set_targ_from_file(char *path, struct sockaddr_in **target)
{
    char line[BUFSIZE];
    FILE *f = NULL;
    int i = 0;
    int allocd = 10;

    f = fopen(path, "r");
    if (!f) fatal("Failed to fopen() the address file!");

    *target = calloc(allocd, sizeof (struct sockaddr_in));

    /* for each line in the file that contains a '.' */
    while ( fgets(line, sizeof (line), f) != NULL ) {
        if (strchr(line, '.') != NULL) {    /* more likely to be a valid IP or DNS addr */
            parse_str_to_targ(line, *target + i++);

            /* if we use up our available address slots, get some more */
            if (i >= allocd) {
                allocd += 10;
                *target = realloc(*target, (sizeof (struct sockaddr_in)) * allocd);
            }
        }
    }

    fclose(f);
    return i;
}

/**
 * Set a single target to the targets list.
 *
 * @param addr IPv4 address in string format
 * @param port TCP port to use
 * @param target Pointer to the targets array to populate
 * @return The number of targets we set (always 1)
 */
int set_single_targ(char *addr, uint16_t port, struct sockaddr_in **target)
{
    *target = calloc(1, sizeof (struct sockaddr_in));
    if (*target == NULL) fatal("calloc() failed: %s", strerror(errno));

    (*target)->sin_family = AF_INET;
    (*target)->sin_port = htons(port);
    resolve_to_ip(addr, *target);

    return 1;
}

