/********************************************************************
 * @file    util.c
 * @brief   Various utility functions that didn't fit elsewhere
 * @author  $LastChangedBy: rcraven $ (hiccups@cmand.org)
 * @version $LastChangedRevision: 501 $
 * @date    $Date: 2014-05-01 00:28:44 -0700 (Thu, 01 May 2014) $
 *
 * $Id: util.c 501 2014-05-01 07:28:44Z rcraven $
 ********************************************************************/
#include "hiccups.h"

#ifndef _WIN
    const char *color_str[] = ANSI_COLOR_STRINGS;
#endif

/**
 * Prints a range of data in hex.
 *
 * @param buf Start of the data range
 * @param len Length of the data range
 * @param brk Number of bytes to print per line
 * @param tabs Number of tabs to insert at the beginning of each new line
 */
void print_binary(const unsigned char *buf, int len, int brk, int tabs)
{
    int i, j;
    for (i = 0; i < len; i++) {
        if ( (i > 0) && (i % brk == 0) ) {
            printf("\n");
            for (j=0;j<tabs;j++) printf("\t");
        }
        printf("%02X ", buf[i]);
    }
    printf("\n");
}

/**
 * Converts bottom 3 bits to string form.
 *
 * @param num The number to convert
 * @param str A buffer of at least length 4
 */
void int_to_binary(int num, char *str)
{
    char *p;
    int z;

    p = str;
    for (z = 4; z > 0; z >>= 1)
        *p++ = (((num & z) == z) ? '1' : '0');
    *p = '\0';
}

/**
 * Prints each header (net, trnspt, data) of a packet in hex.
 *
 * @param packet Start of the buffer containing the packet
 * @param tot_len Total length of the packet buffer
 */
void print_packet(const unsigned char *packet, int tot_len)
{
    struct ip *ip;
    struct tcphdr *tcp;
    int iph_len, tcph_len, hdrs_len;

    ip = (struct ip *) packet;
    iph_len = ip->ip_hl << 2;
    
    tcp = (struct tcphdr *) (packet + iph_len);
    tcph_len = tcp->th_off << 2;
    
    hdrs_len = iph_len + tcph_len;

    printf("\tIP Header:\t");
    print_binary((unsigned char *) ip, iph_len, 10, 3);
    
    printf("\tTCP Header:\t");
    print_binary((unsigned char *) tcp, tcph_len, 10, 3);

    /* only print a data header if the packet has data */
    if (tot_len > hdrs_len) {
        printf("\tData:\t\t");
        print_binary(packet + hdrs_len, tot_len - hdrs_len, 10, 3);
    }
}

/**
 * Wraps printf to print to the terminal in color
 *
 * @param color The color code to use (see hiccups.h)
 * @param format Typical printf stuff
 */
void color_printf(int color, const char *format, ...)
{
    va_list args;
    va_start(args, format);

#ifndef _WIN
    printf("%s", color_str[color]);
    vprintf(format, args);
    printf("%s", color_str[TERM_NONE]);
#else
    HANDLE hStdout;
    hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
    SetConsoleTextAttribute(hStdout, color);
    vprintf(format, args);
    SetConsoleTextAttribute(hStdout, TERM_NONE);
#endif

    va_end(args);
}

/**
 * Trims whitespace from each end of a string, in place
 *
 * @param s The string to trim
 * @return An advanced ptr into s (can't free() this ptr)
 */
char *trim(char *s)
{
    char *end = s + strlen(s) - 1;

    /* trim right */
    while (isspace(*end)) {
        if (end < s) break;
        end--;
    }
    *(end + 1) = '\0';

    /* trim left */
    while (isspace(*s))
        s++;

    return s;
}

/**
 * Reads up to 8 random bytes from /dev/urandom.
 *
 * @return Specified number of random bytes (up to 8)
 */
uint64_t get_rand_bytes(int bytes)
{
    uint64_t rand;
    if (bytes > 8) bytes = 8;

#ifndef _WIN
    int r;

    /* read random bytes from /dev/urandom */
    if ((r = open("/dev/urandom", O_RDONLY)) < 0)
        fatal("** open(\"/dev/urandom\") failed: %s", strerror(errno));
    read(r, &rand, bytes);
    close(r);

#else
    HCRYPTPROV hProvider = 0;

    if (! CryptAcquireContextW(&hProvider, 0, 0, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT | CRYPT_SILENT))
        fatal("** Couldn't init Crypto API for random number generation");

    if (! CryptGenRandom(hProvider, bytes, (BYTE *) &rand)) {
        CryptReleaseContext(hProvider, 0);
        fatal("** Crypto API random number generation failed");
    }

    CryptReleaseContext(hProvider, 0);
#endif

    return rand;
}

/**
 * Stop execution if we don't have root privileges.
 */
void ensure_root_priv(void)
{
    int have_root;

#ifndef _WIN
    have_root = (getuid() == 0) || (geteuid() == 0);
#else
    SID_IDENTIFIER_AUTHORITY NtAuthority = SECURITY_NT_AUTHORITY;
    PSID AdministratorsGroup;
    have_root = AllocateAndInitializeSid(
                &NtAuthority,
                2,
                SECURITY_BUILTIN_DOMAIN_RID,
                DOMAIN_ALIAS_RID_ADMINS,
                0, 0, 0, 0, 0, 0,
                &AdministratorsGroup);

    if (have_root) {
        if (!CheckTokenMembership(NULL, AdministratorsGroup, &have_root)) have_root = 0;
        FreeSid(AdministratorsGroup);
    }
#endif

    if (!have_root)
        fatal("This program must be run as root to access raw sockets!");
}

/**
 * Read data from a file (used for AppSalt).
 *
 * @param path The path to the data file that is to be read
 * @param dest Where to store the bytes that are read
 * @return The length of the content read from the file
 */
int read_data_file(char *path, char **dest)
{
    char buf[BUFSIZE];
    FILE *f = NULL;
    int noms, eaten = 0;

    f = fopen(path, "rb");
    if (!f) fatal("Failed to fopen() the data file!");

    *dest = malloc(BUFSIZE);

    while ( ! feof(f) ) {
        noms = fread(buf, 1, BUFSIZE, f);
        if (noms < 0) fatal("Failed to fread() the data file!");

        memcpy(*dest + eaten, buf, noms);
        eaten += noms;

        *dest = realloc(*dest, eaten + BUFSIZE);
    }
    return eaten;
}

/**
 * A more secure gets().
 *
 * @param buf Where to store the user's input string
 * @param len The maximum length of input to accept from user
 * @return The string length of the user's input
 */
int get_line(char *buf, size_t len)
{
    char *s;

    s = fgets(buf, len, stdin);
    if (s == NULL) perror("Error while reading from stdin");

    if (buf[strlen(buf)-1] != '\n')
        warn("Had to truncate; input string was too long");
    else
        buf[strlen(buf)-1] = '\0'; /* strip trailing newline */

    return strlen(buf);
}

